# code copied from https://github.com/desihub/desimeter/blob/0.6.7/py/desimeter/transform/zhaoburge.py
# with permission from  https://github.com/desihub/desimeter/blob/0.6.7/LICENSE.rst
# and help from Stephen Bailey, Julien Guy, Michael Lampton et al.

# code extracted from DESI-3798 IterFitter8.py


#-------------------IMPORT ZONE-----------------------

import numpy as np

#-----------ZERNIKE NORMALIZED EDGE=1 PACKAGE-----------
#-----------Using BORN-WOLF {n,m} definitions-----------

def factorial(n):
    if n > 1:
       return int(n*factorial(n-1))
    else:
       return 1

def convertNolltoBW(noll):
    # converts a Noll Zernike index to the B&W {n,m} doublet
    n = int(-0.5+np.sqrt(2*noll-1.75))
    m = 0
    base = int(0.5*n**2 +0.5*n + 1)
    diff = noll - base
    if n%2==0:
        m = 2*int(0.5*diff + 0.7)
    else:
        m = 2*int(0.5*diff + 1.2) - 1
    if noll%2>0:
        m = -m
    return np.array([n, m])

def convertWyanttoBW(wyant):
    # converts a Wyant Zernike index to the B&W {n,m}
    halfsum = int(np.sqrt(wyant))
    idif = int(wyant - halfsum**2)
    halfdif = int (idif/2)
    n = halfsum + halfdif
    m = halfsum - halfdif
    if idif%2 >0:
        m = -m
    return np.array([n, m])

def getZernFuncXY(nm, xnorm, ynorm):   # nm is the BornWolf index; BIG NINE:  #1
    # Here, xnorm and ynorm must lie within the unit circle
    rnorm = np.sqrt(xnorm*xnorm + ynorm*ynorm)
    angle = np.arctan2(ynorm,xnorm)
    return getZernRadial(nm,rnorm) * getZernAngular(nm,angle)

def getZernRadial(nm, rnorm):    # BIG NINE: #4
    n = nm[0]             # B&W
    m = np.abs(nm[1])     # B&W
    halfsum = (n+m)/2
    idif = n-m
    halfdif = int(idif/2)
    # n = halfsum + halfdif   # or, halfsum = (n+m)/2
    # m = halfsum - halfdif   # or, halfdif = (n-m)/2
    # loop through the polynomial
    result = 0.
    for i in range(0, halfdif+1):
        expon = int(n-2*i)
        sign = 1 if i%2 == 0 else -1
        numer = sign * factorial(n-i)
        denom = factorial(i) * factorial(halfsum-i) * factorial(halfdif-i)
        coef = numer / denom
        term = coef*np.power(rnorm, expon)
        result = result + term
    return result

def getZernAngular(nm, theta):
    m = nm[1]    # B&W
    if m==0:
        return 1.
    if m>0:
        return np.cos(m*theta)
    m = np.abs(m)         # note this abs() function
    return np.sin(m*theta)

def zernFormulaText(nm):           # BIG NINE: #8
    #---generates a text representation of the specified Zernike function--
    n = nm[0]  # B&W
    m = nm[1]  # B&W
    # print 'New zernFormulaText() is using n, m: ', n, m
    needsine = True if m<0 else False
    m = np.abs(m)
    halfsum = (n+m)/2
    idif = n-m
    halfdif = int(idif/2)

    #--first do the radial part----
    nterms = 0
    s = ''
    #--evaluate the radial polynomial-----
    for i in range(0, halfdif+1):
        nterms = nterms + 1
        # print "Starting with n,  m, i, nterms = ", n, m, i, nterms
        expon = int(n-2*i)                # start with highest exponent
        # print "  expon = ", expon
        sign = 1 if i%2 == 0 else -1      # alternating signs in Zernike series
        numer = sign * factorial(n-i)
        denom = factorial(i) * factorial(halfsum-i) * factorial(halfdif-i)
        coef = numer / denom
        scoef = str(coef)
        if coef==1 and expon>0:          # suppress showing coef=1
            scoef = ''
        if coef > 0 and nterms > 1:
            scoef = '+' + scoef
        s = s + scoef
        if expon > 0:
            s = s + 'r'
        if expon > 1:
            s = s + '^' + str(expon)
    if nterms>1 and m!=0:
        s = '('+ s + ')'

    #--then do the azimuthal part, if any--------
    if m==0:
        return s
    strm = ''
    if m>1:
        strm = str(m)
    if needsine:
        s = s + '*sin(' + strm + 't)'
    else:
        s = s + '*cos(' + strm + 't)'
    return s

#----------End Zernike package with {n,m} B&W indexing---------------------


#-----Zhao-Burge functions built from Zernikes Package---------------------

rh = np.sqrt(0.5)
rt = np.sqrt(2.0)

NCOEFS = 33  # available item coefs

#--------Zernike-Noll Renormalization for AreaIntegral = Pi----------------

squares = np.array([0, 1, 4, 4, 3, 6, 6, 8, 8, 8, 8,  5, 10, 10, 10, 10, 12, 12, 12, 12, 12, 12, 7, 14, 14, 14, 14, 14, 14, 16, 16, 16, 16, 16, 16, 16, 16, 9])
# for Noll index =  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,12, 13, 14, 15, 16, 17, 18, 19, 20, 21,22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37

def normalizeArea(noll):
    # converts an edge=1 Zernike to an area=Pi normalized Zernike
    return np.sqrt(squares[noll])

def getZ(noll, x, y):
    return normalizeArea(noll) * getZernFuncXY(convertNolltoBW(noll), x, y)

def getZhaoBurgeXY(polids, coeffs, x, y, normFactor=1):
    """
    Args:
        coeffs: array of coefficients
        x,y : locations at which to evaluate
    returns dx, dy arrays
    """
    x = x / normFactor
    y = y / normFactor

    dx = np.zeros(len(x))
    dy = np.zeros(len(y))
    for polid, coeff in zip(polids,coeffs):
        zbx, zby, _ = getZhaoBurgeTerm(polid, x, y)
        dx += coeff*zbx
        dy += coeff*zby

    return dx*normFactor, dy*normFactor

def getZhaoBurgeTerm(polid, x, y):

    # Cartesian input x,y; cartesian output x, y, and label.
    # Calls getZ() and delivers area normalized ZB terms
    if polid==0:   # case "S2"  r^0, keep; X translate;
        result =  getZ(1,x,y), 0.0,            'S2'   # parm=0/13 translation
    elif polid==1:   # case "S3"  r^0, keep; Y translate;
        result =  0.0, getZ(1,x,y),             'S3'   # parm=1/13 translation
    elif polid==2:   # case "S4"  magnify
        result =  rh*getZ(2,x,y), rh*getZ(3,x,y), 'S4'  # parm=2/13 dilatation
    elif polid==3:   # case "S5"
        result =  rh*getZ(3,x,y), rh*getZ(2,x,y), 'S5'  # n/a
    elif polid==4:   # case "S6"
        result =  rh*getZ(2,x,y), -rh*getZ(3,x,y), 'S6'  # n/a
    elif polid==5:   # case "S7"
        result =  0.5*getZ(5,x,y), rh*getZ(4,x,y)-0.5*getZ(6,x,y), 'S7' # parm=3/13
    elif polid==6:   # case "S8"
        result =  rh*getZ(4,x,y)+0.5*getZ(6,x,y), 0.5*getZ(5,x,y), 'S8'  # parm=4/13
    elif polid==7:   # case "S9"
        result =  rh*getZ(5,x,y), rh*getZ(6,x,y), 'S9'      # parm=5/13 but TREFOIL: DROPPING THIS
    elif polid==8:   # case "S10"
        result =  rh*getZ(6,x,y), -rh*getZ(5,x,y), 'S10'    # n/a but is SISTER TO S9 !!!
    elif polid==9:   # case "S11"
        result =  rh*getZ(8,x,y), rh*getZ(7,x,y), 'S11'   # parm=6/13
    elif polid==10:  # case "S12"
        result =  0.5*getZ(8,x,y)+0.5*getZ(10,x,y), -0.5*getZ(7,x,y)+0.5*getZ(9,x,y), 'S12' # n/a
    elif polid==11:  # case "S13"
        result =  0.5*getZ(7,x,y)+0.5*getZ(9,x,y), 0.5*getZ(8,x,y)-0.5*getZ(10,x,y), 'S13'  # n/a
    elif polid==12:  # case "S14"
        result =  rh*getZ(10,x,y), -rh*getZ(9,x,y),  'S14'   # n/a
    elif polid==13:  # case "S15"
        result =  rh*getZ(9,x,y), rh*getZ(10,x,y),  'S15'   # n/a
    elif polid==14:  # Case "S16"
        result =  rh*getZ(11,x,y)+0.5*getZ(12,x,y), 0.5*getZ(13,x,y), 'S16'  # n/a; TYPO! 13 not 3
    elif polid==15:  # Case "S17"
        result =  0.5*getZ(13,x,y), rh*getZ(11,x,y)-0.5*getZ(12,x,y), 'S17'  # parm=7/13  TYPO! 13 not 3
    elif polid==16:  # Case "S18"
        result =  0.5*getZ(12,x,y)+0.5*getZ(14,x,y), 0.5*getZ(15,x,y)-0.5*getZ(13,x,y), 'S18' # n/a
    elif polid==17:  # Case "S19"
        result =  0.5*getZ(13,x,y)+0.5*getZ(15,x,y), 0.5*getZ(12,x,y)-0.5*getZ(14,x,y), 'S19'  # n/a
    elif polid==18:  # Case "S20"
        result =  rh*getZ(14,x,y), -rh*getZ(15,x,y), 'S20'  # n/a
    elif polid==19:  # Case "S21"
        result =  rh*getZ(15,x,y), rh*getZ(14,x,y), 'S21', # n/a
    elif polid==20:  # Case "S22"
        result =  rh*getZ(16,x,y), rh*getZ(17,x,y), 'S22'  # parm=8/13
    elif polid==21:  # Case "S23"
        result =  0.5*getZ(17,x,y)+0.5*getZ(19,x,y), 0.5*getZ(16,x,y)-0.5*getZ(18,x,y), 'S23'  # n/a
    elif polid==22:  # Case "S24"
        result =  0.5*getZ(16,x,y)+0.5*getZ(18,x,y), -0.5*getZ(17,x,y)+0.5*getZ(19,x,y), 'S24' # n/a
    elif polid==23:  # Case "S25"
        result =  0.5*getZ(19,x,y)+0.5*getZ(21,x,y), 0.5*getZ(18,x,y)-0.5*getZ(20,x,y), 'S25'  # n/a
    elif polid==24:  # Case "S26"
        result =  0.5*getZ(18,x,y)+0.5*getZ(20,x,y), -0.5*getZ(19,x,y)+0.5*getZ(21,x,y), 'S26'  # n/a
    elif polid==25:  # Case "S27"
        result =  rh*getZ(21,x,y), rh*getZ(20,x,y), 'S27'                 # n/a
    elif polid==26:  # Case "S28"
        result =  rh*getZ(20,x,y), -rh*getZ(21,x,y), 'S28'                # n/a
    elif polid==27:  #  case "T4"
        result =  rh*getZ(3,x,y), -rh*getZ(2,x,y), 'T4'                   # parm=9/13 rotation
    elif polid==28:  # case "T7"
        result =  rh*getZ(4,x,y)-0.5*getZ(6,x,y), -0.5*getZ(5,x,y), 'T7'  # parm=10/13
    elif polid==29:  # case "T8"
        result =  0.5*getZ(5,x,y), -rh*getZ(4,x,y)-0.5*getZ(6,x,y), 'T8'  # parm=11/13; TWO MINUS SIGNS
    elif polid==30:  # case "T11"
        result =  rh*getZ(7,x,y), -rh*getZ(8,x,y), 'T11'                  # parm=12/13
    elif polid==31:  # case "T12"
        result =  -0.5*getZ(7,x,y)+0.5*getZ(9,x,y), -0.5*getZ(8,x,y)-0.5*getZ(10,x,y), 'T12' # n/a
    elif polid==32:  # case "T13"
        result =  0.5*getZ(8,x,y)-0.5*getZ(10,x,y), -0.5*getZ(7,x,y)-0.5*getZ(9,x,y), 'T13' # n/a
    else:
        print("ZhaoBurgeTerm() is exitting because unsatisfied polid = ", polid)
        quit()
    return result

#---------end of Zhao-Burge evaluator------------------------


#- Transformation and fit routines

from scipy.optimize import minimize

#- Tranformation in reduced coordinates space, to be used by minimizer
def transform(x, y, scale, rotation, offset_x, offset_y, zbpolids=None, zbcoeffs=None):
    """
    TODO: document
    """
    x = x + offset_x
    y = y + offset_y
    xx = (x*np.cos(rotation) - y*np.sin(rotation))*scale # + offset_x
    yy = (x*np.sin(rotation) + y*np.cos(rotation))*scale # + offset_y

    if zbpolids is not None:
        if zbcoeffs is None:
            raise RuntimeError("need none or both zppolids and zbcoeffs")
        dx, dy = getZhaoBurgeXY(zbpolids,zbcoeffs, xx, yy)
        xx += dx
        yy += dy

    return xx, yy

def fit_scale_rotation_offset(x, y, xp, yp, fitzb=False, zbpolids=None, zbcoeffs=None, fixed_scale=None, fixed_rotation=None):
    """
    Fit scale, rotation, offset plus optional Zhao-Burge corrections
    for x,y -> xp,yp.
    TODO: document details
    """


    # if coeffs are not Null we have to apply the ZB correction in
    # the fit of the coefficients scale, rotation, offset_x, offset_y
    # because values of non-linear terms are changing with
    # the scale, rotation, offset_x, offset_y that is applied to
    # the input coordinates.

    # however the first if just a fit of scale, rotation, offset_x, offset_y

    def func(params, x, y, xp, yp):
        scale, rotation, offset_x, offset_y = params[0:4]
        if fixed_scale is not None: scale=fixed_scale # fixed value
        if fixed_rotation is not None: rotation=fixed_rotation # fixed value
        xx, yy = transform(x, y, scale, rotation, offset_x, offset_y, zbpolids=zbpolids, zbcoeffs=zbcoeffs)
        dr2 = np.sum((xx-xp)**2 + (yy-yp)**2)
        if fixed_scale is not None: dr2 += 1e6*(scale-fixed_scale)**2 # prior on fixed value
        if fixed_rotation is not None: dr2 += 1e6*(rotation-fixed_rotation)**2 # prior on fixed value
        return dr2

    p0 = np.array([1.0,0.0,0.0,0.0])
    p  = minimize(func, p0, args=(x, y, xp, yp))

    scale, rotation, offset_x, offset_y = p.x
    if fixed_scale is not None: scale=fixed_scale # fixed value
    if fixed_rotation is not None: rotation=fixed_rotation # fixed value

    if fitzb :

        # here we refit only the ZB coefficients, but those should include provision
        # for adjustment of shift translation rotation
        xx, yy = transform(x, y, scale, rotation, offset_x, offset_y)
        zbpolids, zbcoeffs = fitZhaoBurge(xx, yy, xp, yp, zbpolids)
        return scale, rotation, offset_x, offset_y, zbpolids, zbcoeffs
    else:
        return scale, rotation, offset_x, offset_y


def fitZhaoBurge(x, y, xp, yp, polids=None, normFactor=1):
    x = x / normFactor
    y = y / normFactor
    xp = xp / normFactor
    yp = yp / normFactor

    dx = xp-x
    dy = yp-y

    nx = len(x)

    # here we choose the polynomials
    # 0  = S2 = translation along X
    # 1  = S3 = translation along Y
    # 2  = S4 = magnification
    # 27 = T4 = rotation

    if polids is None :
        polids = np.array([0,1,2,3,4,5,6,9,20,27,28,29,30],dtype=int)

    nzb = polids.size
    H = np.zeros((2*nx, nzb))
    for i,polid in enumerate(polids) :
        zbx, zby, _ = getZhaoBurgeTerm(polid, x, y)
        H[0:nx, i] = zbx
        H[nx:, i] = zby

    # print("H shape", H.shape)

    A = H.T.dot(H)
    b = H.T.dot(np.concatenate([dx, dy]))
    # print("shapes", A.shape, b.shape)
    coeffs = np.linalg.solve(A, b)

    return polids, coeffs


import numpy

from .exceptions import CoordIOError


def unitDiskify(x, y, scaleR=None):
    """
    Convert x y to unit disk, return the radial scale factor

    # note scaleR is the max radius of the disk in original units
    # divide by this to put things on a unit disk...
    """
    r = numpy.sqrt(x**2 + y**2)
    theta = numpy.arctan2(y, x)
    if scaleR is None:
        scaleR = numpy.max(r)
    _r = r / scaleR
    _x = numpy.cos(theta) * _r
    _y = numpy.sin(theta) * _r
    return _x, _y, scaleR


def unDiskify(x, y, scaleR):
    """
    Convert x y to unit disk, return the radial scale factor
    """
    r = numpy.sqrt(x**2 + y**2) * scaleR
    theta = numpy.arctan2(y, x)

    _x = numpy.cos(theta) * r
    _y = numpy.sin(theta) * r
    return _x, _y


def _orthoZern(x, y, zernOrder=20):
    """
    Return zernike gradient in x and y, zernOrder is the order of the polynomial.


    transcribed from: https://doi.org/10.1364/JOSAA.35.000840
    typo was fixed by author of paper (s#17 and a few more were wrong)
    but private correspondence fixed that.  Orthonormalized zernike polynomials
    in x/y.

    Parameters
    -----------
    x : float
        unit disk normalized x coordinate
    y: float
        unit disk normalized y coordinate
    zernOrder : int
        number of zernike coefficients to return in range (1-20)

    Returns
    ---------
    result : 2 element numpy.ndarray
        xy vector
    """

    out = numpy.array([
        [1, 0],  # s2
        [0, 1],  # s3
        numpy.sqrt(3) * numpy.array([x, y]),  # s4
        numpy.sqrt(3) * numpy.array([y, x]),  # s5
        numpy.sqrt(3) * numpy.array([x, -y]),  # s6
        numpy.array([6 * x * y, 3 * x**2 + 9 * y**2 - 2]) / numpy.sqrt(3), # s7
        numpy.array([9 * x**2 + 3 * y**2 - 2, 6 * x * y]) / numpy.sqrt(3), # s8

        # s9
        numpy.array([
            12 * x * y,
            6 * x**2 - 12 * y**2 + 1
        ]) / (2 * numpy.sqrt(2)),

        # s10
        numpy.array([12 * x**2 - 6 * y**2 - 1, -12 * x * y]) / numpy.sqrt(8),

        # s11
        numpy.sqrt(21 / 62) * numpy.array([x, y]) * (15 * x**2 + 15 * y**2 - 7),

        # s12
        numpy.sqrt(7) * numpy.array([
            x * (10 * x**2 - 3),
            y * (3 - 10 * y**2)
        ]) / 2,

        # s13
        numpy.sqrt(21 / 38) * numpy.array([
            x * (15 * x**2 + 5 * y**2 - 4),
            y * (5 * x**2 + 15 * y**2 - 4)
        ]),

        # s14
        numpy.array([
            x * (35 * x**2 - 27 * y**2 - 6),
            y * (-27 * y**2 + 35 * x**2 - 6)
        ]) / numpy.sqrt(5/62),

        # s15
        numpy.sqrt(35 / 3) * numpy.array([
            y * (3 * x**2 - y**2),
            x * (x**2 - 3 * y**2)
        ]),

        # s16
        numpy.array([
            315 * (x**2 + y**2) * (5 * x**2 + y**2) - 30 * (33 * x**2 + 13 * y**2) + 83,
            60 * x * y * (21 * (x**2 + y**2) - 13)
        ]) / (2 * numpy.sqrt(1077)),

        # s17
        numpy.array([
            60 * x * y * (21 * (x**2 + y**2) - 13),
            315 * (x**2 + y**2) * (x**2 + 5 * y**2) - 30 * (13 * x**2 + 33 * y**2) + 83
        ]) / (2 * numpy.sqrt(1077)),

        # s18
        3 * numpy.array([
            140 * (860 * x**4 - 45 * x**2 * y**2 - 187 * y**4) - 30 * (1685 * x**2 - 522 * y**2) + 1279,
            -40 * x * y * (105 * x**2 + 2618 * y**2 - 783)
        ]) / (2 * numpy.sqrt(2489214)),

        # s19
        3 * numpy.array([
            40 * x * y * (2618 * x**2 + 105 * y**2 - 783),
            140 * (187 * x**4 + 45 * x**2 * y**2 - 860 * y**4) - 30 * (522 * x**2 - 1685 * y**2) - 1279
        ]) / (2 * numpy.sqrt(2489214)),

        #  s20
        (1 / 16) * numpy.sqrt(7 / 13557143) * numpy.array([
            60 * (10948 * x**4 - 7830 * x**2 * y**2 + 2135 * y**4 - 3387 * x**2 - 350 * y**2) + 11171,
            -1200 * x * y * (261 * x**2 - 427 * y**2 + 35)
        ]),

        # s21
        (1 / 16) * numpy.sqrt(7 / 13557143) * numpy.array([
            1200 * x * y * (427 * x**2 - 261 * y**2 + 35),
            60 * (2135 * x**4 - 7830 * x**2 * y**2 + 10948 * y**4 - 350 * x**2 - 3387 * y**2) + 11171
        ])
    ])

    # note computing everything up to 20 (the max coeffs provided)
    # but only returning the order requested
    out = out[:zernOrder]

    return out


def orthoZern(x, y, zernOrder=20):
    """
    Compute the orthonormailzed zernike gradients in x and y,
    zernOrder is the order of the polynomial.


    transcribed from: https://doi.org/10.1364/JOSAA.35.000840
    typo was fixed by author of paper (s#17 and a few more were wrong)
    but private correspondence fixed that.  Orthonormalized zernike polynomials
    in x/y.

    Parameters
    -----------
    x : numpy.ndarray
        unit disk normalized x coordinate
    y: numpy.ndarray
        unit disk normalized y coordinate
    zernOrder : int
        number of zernike coefficients to return in range (1-20)

    Returns
    ---------
    dUdx : numpy.ndarray
        gradient in x direction, matrix shape n x zernOrder
    dUdy : numpy.ndarray
        gradient in y direction, matrix shape n x zernOrder
    """

    maxRad = numpy.max(numpy.sqrt(x**2 + y**2))
    if maxRad > 1:
        raise CoordIOError("x y coords must be on unit disk")

    if zernOrder > 20:
        raise CoordIOError("maximum zernike order is 20")

    nPts = len(x)
    dx = numpy.zeros((nPts, zernOrder))
    dy = numpy.zeros((nPts, zernOrder))
    for ii, (_x, _y) in enumerate(zip(x, y)):
        dxy = _orthoZern(_x, _y, zernOrder)
        dx[ii, :] = dxy[:, 0]
        dy[ii, :] = dxy[:, 1]

    return dx, dy


def gradZern(x, y, zernOrder=20):
    """
    Computes zernike polynomial value and gradients in x/y.

    This routine isn't orthonormalized so seems to perform not quite as well
    as the orthoZern routine.  However this routine allows one to fit an
    arbitrarily large number of coefficients, while the other routine only
    allows 20.

    from https://doi.org/10.1364/OE.26.018878
     Pseudo-code to calculate unit-normalized Zernike polynomials and their x,y derivatives

       Numbering scheme:
       Within a radial order, sine terms come first
               ...
             sin((n-2m)*theta)   for m = 0,..., [(n+1)/2]-1
               ...
                1                for n even, m = n/2
               ...
             cos((n-2m)*theta)   for m = [n/2]+1,...,n
               ...

       INPUT:
       x, y normalized (x,y) coordinates in unit circle
       zernOrder: Maximum Zernike radial order

       OUTPUT:
       Zern[...]   array to receive value of each Zernike polynomium at (x,y)
       dUdx[...]   array to receive each derivative dU/dx at (x,y)
       dUdy[...]   array to receive each derivative dU/dy at (x,y)

    Parameters
    -----------
    x : float or numpy.ndarray
        unit disk normalized x coordinate
    y: float or numpy.ndarray
        unit disk normalized y coordinate
    zernOrder : int
        number of radial orders (total orders will be greater!)

    Returns
    --------
    dUdx : numpy.ndarray
        gradient in x direction, matrix shape n x totalOrder
        where totalOrder > zernOrder
    dUdy : numpy.ndarray
        gradient in y direction, matrix shape n x totalOrder
        where totalOrder > zernOrder

    """
    maxRad = numpy.max(numpy.sqrt(x**2 + y**2))
    if maxRad > 1:
        raise CoordIOError("x y coords must be on unit disk")

    nTerms = numpy.sum(numpy.arange(zernOrder + 2)) + 1
    Zern = numpy.zeros(nTerms)
    dUdx = numpy.zeros(nTerms)
    dUdy = numpy.zeros(nTerms)

    if hasattr(x, "__len__"):
        # x, y are vectors
        Zern = numpy.array([Zern] * len(x)).T
        dUdx = numpy.array([dUdx] * len(x)).T
        dUdy = numpy.array([dUdy] * len(x)).T

    # double x, y, Zern[*], dUdx[*], dUdy[*]

    # int nn, mm, kndx, jbeg, jend, jndx, even, nn1, nn2
    # int jndx1, jndx11, jndx2, jndx21
    # double pval, qval

    # pseudocode is 1 indexed, ugh, stick with it
    # and modify later

    Zern[1] = 1.                                  # (0,0)
    dUdx[1] = 0.                                  # (0,0)
    dUdy[1] = 0.                                  # (0,0)

    Zern[2] = y                                   # (1,0)
    Zern[3] = x                                   # (1,1)
    dUdx[2] = 0.                                  # (1,0)
    dUdx[3] = 1.                                  # (1,1)
    dUdy[2] = 1.                                  # (1,0)
    dUdy[3] = 0.                                  # (1,1)

    kndx = 1                # index for term from 2 orders down
    jbeg = 2                # start index for current radial order
    jend = 3                # end index for current radial order
    jndx = 3                # running index for current Zern
    even = -1
    #  Outer loop in radial order index
    for nn in range(2, zernOrder + 1):  # 1 indexed
        even = -even          # parity of radial index
        jndx1 = jbeg           # index for 1st ascending series in x
        jndx2 = jend           # index for 1st descending series in y
        jndx11 = jndx1 - 1      # index for 2nd ascending series in x
        jndx21 = jndx2 + 1      # index for 2nd descending series in y
        jbeg = jend + 1       # end of previous radial order +1
        nn2 = nn // 2
        nn1 = (nn - 1) // 2
        #  Inner loop in azimuthal index
        for mm in range(0, nn + 1):  # 1 indexed
            jndx += 1                  # increment running index for current Zern

            if (mm == 0):
                Zern[jndx] = x * Zern[jndx1] + y * Zern[jndx2]
                dUdx[jndx] = Zern[jndx1] * nn
                dUdy[jndx] = Zern[jndx2] * nn

            elif (mm == nn):
                Zern[jndx] = x * Zern[jndx11] - y * Zern[jndx21]
                dUdx[jndx] = Zern[jndx11] * nn
                dUdy[jndx] = -Zern[jndx21] * nn

            elif ((even > 0) and (mm == nn2)):              # logical “AND”
                Zern[jndx] = 2. * (x * Zern[jndx1] + y * Zern[jndx2]) - Zern[kndx]
                dUdx[jndx] = 2. * nn * Zern[jndx1] + dUdx[kndx]
                dUdy[jndx] = 2. * nn * Zern[jndx2] + dUdy[kndx]
                kndx += 1                        # increment kndx

            elif ((even < 0) and (mm == nn1)):              # logical “AND”
                qval = Zern[jndx2] - Zern[jndx21]
                Zern[jndx] = x * Zern[jndx11] + y * qval - Zern[kndx]
                dUdx[jndx] = Zern[jndx11] * nn + dUdx[kndx]
                dUdy[jndx] = qval * nn + dUdy[kndx]
                kndx += 1                        # increment kndx

            elif ((even < 0) and (mm == nn1 + 1)):            # logical “AND”
                pval = Zern[jndx1] + Zern[jndx11]
                Zern[jndx] = x * pval + y * Zern[jndx2] - Zern[kndx]
                dUdx[jndx] = pval * nn + dUdx[kndx]
                dUdy[jndx] = Zern[jndx2] * nn + dUdy[kndx]
                kndx += 1                        # increment kndx

            else:
                pval = Zern[jndx1] + Zern[jndx11]
                qval = Zern[jndx2] - Zern[jndx21]
                Zern[jndx] = x * pval + y * qval - Zern[kndx]
                dUdx[jndx] = pval * nn + dUdx[kndx]
                dUdy[jndx] = qval * nn + dUdy[kndx]
                kndx += 1                        # increment kndx

            jndx11 = jndx1                   # update indices
            jndx1 += 1
            jndx21 = jndx2
            jndx2 -= 1
            # End of inner azimuthal loop

        jend = jndx
        # print("jend", jend)
        # End of outer radial order loop

    # throw out first term (zero indexing didn't populate it)
    Zern = Zern[1:]  # not returned but computed if we want it later?
    dUdx = dUdx[1:]
    dUdy = dUdy[1:]

    return dUdx.T, dUdy.T


class ZernFit(object):
    def __init__(self, xMeas, yMeas, xExpect,
                 yExpect, orders=20, method="ortho", scaleR=None):
        """
        Basic least squares fitter for zernike gradients.  Translation,
        rotation and scale are expected to be removed from x/y Meas prior to
        zernike fitting

        Parameters
        ------------
        xMeas : numpy.ndarray
            x measurements
        yMeas : numpy.ndarray
            y measurements
        xExpect : numpy.ndarray
            expected (true) x coordinates
        yExpect : numpy.ndarray
            expected (true) y coordinates
        orders : int
            number of orders to include in fit. If method == "grad" these are
            radial order of zernike's, so the total number of coeffs will be >
            input orders.
        method : str
            either "ortho" or "grad".  Ortho uses method from:
            https://doi.org/10.1364/JOSAA.35.000840
            Grad uses method from:
            https://doi.org/10.1364/OE.26.018878
        scaleR : float or None
            if not None, apply this radial scale to xyMeas input
            to put coords on unit disk
        """
        self.orders = orders
        self.xMeas = xMeas
        self.yMeas = yMeas
        self.xExpect = xExpect
        self.yExpect = yExpect
        self.xErr = xExpect - xMeas
        self.yErr = yExpect - yMeas

        # force all measurements to unit disk
        # scale expected locations by same factor
        self.xMeasUnit, self.yMeasUnit, self.rScale = unitDiskify(
            xMeas, yMeas, scaleR
        )
        self.xExpectUnit, self.yExpectUnit, _junk = unitDiskify(
            xExpect, yExpect, self.rScale
        )
        self.xErrUnit = self.xExpectUnit - self.xMeasUnit
        self.yErrUnit = self.yExpectUnit - self.yMeasUnit

        if method == "ortho":
            self._zernFunc = orthoZern
        elif method == "grad":
            self._zernFunc = gradZern
        else:
            raise CoordIOError("method must be either ortho or grad")

        # zx/zy are shape n x orders
        self.zx, self.zy = self._zernFunc(
            self.xMeasUnit, self.yMeasUnit, self.orders
        )

        # stack them and do the fit
        self.zxyStack = numpy.vstack((self.zx, self.zy))
        self.exyStack = numpy.hstack((self.xErrUnit, self.yErrUnit))

        # solve Ax=b for x (the coefficients)
        self.coeff, resid, rank, s = numpy.linalg.lstsq(
            self.zxyStack, self.exyStack
        )

    def apply(self, xMeas, yMeas):
        # apply the fit to measured data
        xMeasUnit, yMeasUnit, _junk = unitDiskify(xMeas, yMeas, self.rScale)
        zx, zy = self._zernFunc(xMeasUnit, yMeasUnit, self.orders)
        dx = zx @ self.coeff
        dy = zy @ self.coeff

        xFitUnit = xMeasUnit + dx
        yFitUnit = yMeasUnit + dy

        # recale back up from unit disk
        xFit, yFit = unDiskify(xFitUnit, yFitUnit, self.rScale)
        return xFit, yFit





import os

import pandas as pd

from coordio.calibration import Calibration

from .exceptions import CoordIOError


# default/constant values collected here...for now

# a scale factor between xy focal and xy wok
# JSG: setting this to 1.0. We may remove it eventually.
# FOCAL_SCALE = 1.0

# SITE_TO_SCALE cale is focal plane scale mm / wok scale mm
# found via dither analysis
SITE_TO_SCALE = {"LCO": 1.0003, "APO": 0.99982}

MICRONS_PER_MM = 1000
GFA_PIXEL_SIZE = 13.5  # micron
GFA_CHIP_CENTER = 1024  # unbinned pixels

EPOCH = 2451545.0  # J2000
WAVELENGTH = 6231.0  # angstrom, GFA wavelength
VALID_WAVELENGTHS = set([5400.0, 6231.0, 16600.0])

# wavelengths in angstroms
WAVE_TO_INST = {5400.0: "Boss", 6231.0: "GFA", 16600.0: "Apogee"}  # sdss-r band

INST_TO_WAVE = {"Boss": 5400.0, "GFA": 6231.0, "Apogee": 16600.0}

PLATE_SCALE = {"APO": 217.7358, "LCO": 330.275}  # mm/deg

POSITIONER_HEIGHT = 143.1  # mm, distance from wok surface to fiber
ALPHA_LEN = 7.4  # mm, length of the alpha arm (distance between alpha and beta axes)
BETA_LEN = 16.2  # mm, distance from beta axis to top flat edge of the robot
# extracted from ZEMAX model model
# z location of wok vertex in FP coords
APO_WOK_Z_OFFSET_ZEMAX = -776.4791 - POSITIONER_HEIGHT
LCO_WOK_Z_OFFSET_ZEMAX = -993.0665 - POSITIONER_HEIGHT

APO_MAX_FIELD_R = 1.5  # max field radius (deg)
LCO_MAX_FIELD_R = 1.1  # max field radius (deg)

# nominal fiber positions in beta arm frame
MET_BETA_XY = [14.314, 0]
# boss xy position in solid model
BOSS_BETA_XY = [14.965, -0.376]
# apogee xy position in solid model
AP_BETA_XY = [14.965, 0.376]

# default orientation of tangent to a flat wok
IHAT = [0, -1, 0]
JHAT = [1, 0, 0]
KHAT = [0, 0, 1]


def getFPModelParams(site, direction, waveCat):
    """Find the right focal plane model to use given direction and waveCat.

    Parameters
    ----------
    site : str
        The site string.
    direction : str
        "focal" or "field" for the direction of the conversion
    waveCat : str
        "Apogee", "Boss", or "GFA" for the wavelength

    Returns
    ---------
    R: float
        radius of spherical curvature (mm)
    b: float
        z-location of sphere center in focal plane coordinate system
    c0: float
        1st order coeff
    c1: float
        3rd order coeff
    c2: float
        5th order coeff
    c3: float
        7th order coeff
    c4: float
        9th order coeff
    """

    # filter for correct row in the model dataframe
    if direction not in ["focal", "field"]:
        raise CoordIOError(
            "direction must be one of focal or field, got %s" % direction
        )
    if waveCat not in ["Apogee", "Boss", "GFA"]:
        raise CoordIOError(
            "waveCat must be one of Apogee Boss or GFA, got %s" % waveCat
        )

    row = calibration.FP_MODEL.loc[(site, direction, waveCat)]

    R = float(row.R)
    b = float(row.b)
    c0 = float(row.c0)
    c1 = float(row.c1)
    c2 = float(row.c2)
    c3 = float(row.c3)
    c4 = float(row.c4)
    return R, b, c0, c1, c2, c3, c4


def getWokOrient(site):
    """Return the wok orientation from the wok calibration files.

    Returns
    --------
    x : float
        x position of wok vertex in focal plane coords mm
    y : float
        y position of wok vertex in focal plane coords mm
    z : float
        z position of wok vertex in focal plane coords mm
    xTilt : float
        tilt of wok coord sys about focal plane x axis deg
    yTilt : float
        tilt of wok coord sys about focal plane y axis deg
    """

    row = calibration.wokOrient.loc[site]
    x = float(row.x)
    y = float(row.y)
    z = float(row.z)
    xTilt = float(row.xTilt)
    yTilt = float(row.yTilt)

    return x, y, z, xTilt, yTilt


def getHoleOrient(site, holeID):
    """Return orientation of hole position in the wok.

    Parameters
    ----------
    site : str
        The site string.
    holeID : str or list of str
        Either a string with a valid holeID (e.g. R+13C1) or a list of
        valid holeIDs. In the latter case, the returned arrays are
        2D with axis 1 being the holeID dimmension.

    Returns
    --------
    b : numpy.ndarray
        [x,y,z] base position of hole in wok coords (mm)
    iHat : numpy.ndarray
        [x,y,z] unit vector, direction of x Tangent in wok coords
    jHat : numpy.ndarray
        [x,y,z] unit vector, direction of y Tangent in wok coords
    kHat : numpy.ndarray
        [x,y,z] unit vector, direction of z Tangent in wok coords
    """

    wokCoordsH = calibration.wokCoords.loc[(site, holeID)]

    b = wokCoordsH.loc[["xWok", "yWok", "zWok"]].to_numpy()
    iHat = wokCoordsH.loc[["ix", "iy", "iz"]].to_numpy()
    jHat = wokCoordsH.loc[["jx", "jy", "jz"]].to_numpy()
    kHat = wokCoordsH.loc[["kx", "ky", "kz"]].to_numpy()

    return b, iHat, jHat, kHat


def getPositionerData(site, holeID):
    """Return data specific to a positioner.

    Returns:
    --------
    alphaArmLength : float
        alpha arm length (mm)
    metX : float
        metrology fiber x (mm), in beta arm coordinates
    metY : float
        metrology fiber y (mm), in beta arm coordinates
    apX : float
        apogee fiber y (mm), in beta arm coordinates
    apY : float
        apogee fiber x (mm), in beta arm coordinates
    bossX : float
        boss fiber y (mm), in beta arm coordinates
    bossY : float
        boss fiber x (mm), in beta arm coordinates
    alphaOffset : float
        alpha arm offset at (0, 0), in degrees
    betaOffset : float
        beta arm offset at (0, 0), in degrees
    dx: float
        y offset (mm), calibration to capture small displacements of
        tangent y
    dy: float
        z offset (mm), calibration to capture small displacements of
        tangent x

    """

    if holeID not in calibration.VALID_HOLE_IDS:
        raise CoordIOError("%s is not a valid hole ID" % holeID)

    row = calibration.positionerTable.loc[(site, holeID)]

    aal = float(row.alphaArmLen)
    metX = float(row.metX)
    metY = float(row.metY)
    apX = float(row.apX)
    apY = float(row.apY)
    bossX = float(row.bossX)
    bossY = float(row.bossY)
    alphaOffset = float(row.alphaOffset)
    betaOffset = float(row.betaOffset)
    dx = float(row.dx)
    dy = float(row.dy)

    return aal, metX, metY, apX, apY, bossX, bossY, alphaOffset, betaOffset, dx, dy


def parseDesignRef():
    designRefFile = os.path.join(calibration.paths[0], "fps_DesignReference.txt")
    _row = []
    _col = []
    _xWok = []
    _yWok = []
    _fType = []
    _holeName = []

    with open(designRefFile, "r") as f:
        lines = f.readlines()
    for line in lines:
        line = line.strip()
        line = line.split("#")[0]
        if not line:
            continue
        row, col, x, y, fType = line.split()

        col = int(col)
        row = int(row)

        # make col 1 indexed to match pdf maps
        # and wok hole naming convention
        if row == -99:
            holeName = "F%i" % col
            col = None
            row = None

        elif row <= 0:
            holeName = "R%iC%i" % (row, col + 1)
        else:
            holeName = "R+%iC%i" % (row, col + 1)

        _row.append(row)
        _col.append(col)
        _xWok.append(float(x))
        _yWok.append(float(y))
        _fType.append(fType)
        _holeName.append(holeName)

    d = {}
    d["holeName"] = _holeName
    d["row"] = _row
    d["col"] = _col
    d["xWok"] = _xWok
    d["yWok"] = _yWok
    d["fType"] = _fType

    df = pd.DataFrame(d)
    return df


calibration = Calibration()



from coordio import Observed, Site, Field, FocalPlane
from coordio import CoordIOError, CoordIOUserWarning
from coordio.defaults import APO_MAX_FIELD_R, LCO_MAX_FIELD_R
from coordio.defaults import VALID_WAVELENGTHS
import numpy

import pytest

# from sdssconv.fieldCoords import fieldToFocal, focalToField

numpy.random.seed(0)

SMALL_NUM = 1e-10

# from sdssconv.fieldCoords import parallacticAngle
apoSite = Site("APO")
lcoSite = Site("LCO")

# set time (not used but required for Observed coords)
apoSite.set_time(2458863, scale='TAI')
lcoSite.set_time(2458863, scale='TAI')


def test_observedToField_APO():

    #### start checks along the meridian
    ####################################
    azCen = 180 # (pointing south)
    altCen = 45
    obsCen = Observed([[altCen, azCen]], site=apoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=apoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert abs(float(field.x)) < SMALL_NUM
    assert float(field.y) > 0
    assert numpy.abs(field.flatten()[0]-90) < SMALL_NUM
    assert numpy.abs(field.flatten()[1]-1) < SMALL_NUM

    azCen = 0 # (north)
    altCen = 45 # above north star
    obsCen = Observed([[altCen, azCen]], site=apoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=apoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert abs(float(field.x)) < SMALL_NUM
    assert float(field.y) < 0
    assert numpy.abs(field.flatten()[0]-270) < SMALL_NUM
    assert numpy.abs(field.flatten()[1]-1) < SMALL_NUM

    azCen = 0  # (north)
    altCen = 20  # below north star
    obsCen = Observed([[altCen, azCen]], site=apoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=apoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert abs(float(field.x)) < SMALL_NUM
    assert float(field.y) > 0
    assert numpy.abs(field.flatten()[0]-90) < SMALL_NUM
    assert numpy.abs(field.flatten()[1]-1) < SMALL_NUM

    ##### check field rotation (off meridian)
    #########################################
    # remember +x is eastward
    azCen = 180 + 20 # (south-west)
    altCen = 45
    obsCen = Observed([[altCen, azCen]], site=apoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=apoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert float(field.x) > 0
    assert float(field.y) > 0
    assert field.flatten()[0] > 1
    assert field.flatten()[0] < 89

    # check field rotation (off meridian)
    # remember +x is eastward
    azCen = 180 - 20 # (south-east)
    altCen = 45
    obsCen = Observed([[altCen, azCen]], site=apoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=apoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert float(field.x) < 0
    assert float(field.y) > 0
    assert field.flatten()[0] > 91
    assert field.flatten()[0] < 179


    # check field rotation (off meridian)
    # remember +x is eastward
    azCen = 10 # (slightly east of north)
    altCen = apoSite.latitude - 10
    obsCen = Observed([[altCen, azCen]], site=apoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=apoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert float(field.x) < 0
    assert float(field.y) > 0

    azCen = 10 # (slightly east of north)
    altCen = apoSite.latitude + 10
    obsCen = Observed([[altCen, azCen]], site=apoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=apoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert float(field.x) < 0
    assert float(field.y) < 0


def test_observedToField_LCO():

    #### start checks along the meridian
    ####################################
    azCen = 0 # (north)
    altCen = 45
    obsCen = Observed([[altCen, azCen]], site=lcoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=lcoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert abs(float(field.x)) < SMALL_NUM
    assert float(field.y) < 0

    azCen = 180 # (south)
    altCen = numpy.abs(lcoSite.latitude) + 10 # above SCP, (lat is negative!)
    obsCen = Observed([[altCen, azCen]], site=lcoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=lcoSite)
    field = Field(obsCoord, field_center=obsCen)

    assert abs(float(field.x)) < SMALL_NUM
    assert float(field.y) > 0

    azCen = 180 # (south)
    altCen = numpy.abs(lcoSite.latitude) - 10 # below SCP, (lat is negative!)
    obsCen = Observed([[altCen, azCen]], site=lcoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=lcoSite)
    field = Field(obsCoord, field_center=obsCen)
    assert abs(float(field.x)) < SMALL_NUM
    assert float(field.y) < 0

    ##### check field rotation (off meridian)
    #########################################
    # remember +x is eastward
    azCen = 20 # (north-east)
    altCen = 45
    obsCen = Observed([[altCen, azCen]], site=lcoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=lcoSite)
    field = Field(obsCoord, field_center=obsCen)
    assert float(field.x) < 0
    assert float(field.y) < 0

    # remember +x is eastward
    azCen = 290 # (north-west)
    altCen = 45
    obsCen = Observed([[altCen, azCen]], site=lcoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=lcoSite)
    field = Field(obsCoord, field_center=obsCen)
    assert float(field.x) > 0
    assert float(field.y) < 0

    # # check field rotation (off meridian)
    # # remember +x is eastward
    azCen = 170 # (slightly east of south)
    altCen = lcoSite.latitude - 10
    obsCen = Observed([[altCen, azCen]], site=lcoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=lcoSite)
    field = Field(obsCoord, field_center=obsCen)
    assert float(field.x) < 0
    assert float(field.y) < 0

    azCen = 190 # (slightly west of south)
    altCen = lcoSite.latitude - 10
    obsCen = Observed([[altCen, azCen]], site=lcoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=lcoSite)
    field = Field(obsCoord, field_center=obsCen)
    assert float(field.x) > 0
    assert float(field.y) < 0


def test_raises():
    azCen = 190  # (slightly west of south)
    altCen = lcoSite.latitude - 10
    badCen = Observed([[altCen, azCen], [altCen, azCen]], site=lcoSite)
    obsCoord = Observed([[altCen+1, azCen]], site=lcoSite)
    with pytest.raises(CoordIOError):
        Field(obsCoord)
    with pytest.raises(CoordIOError):
        Field(obsCoord, field_center=[1,2])
    with pytest.raises(CoordIOError):
        Field(obsCoord, field_center=badCen)


def test_field_obs_cycle():
    # try a bunch of pointings make sure the round trip works
    azCens = numpy.random.uniform(1,359, size=30)
    altCens = numpy.random.uniform(0,89, size=30)
    for site in [lcoSite, apoSite]:
        for azCen, altCen in zip(azCens, altCens):
            azCoords = azCen + numpy.random.uniform(-1,1, size=10)
            altCoords = altCen + numpy.random.uniform(-1,1, size=10)
            altAzs = numpy.array([altCoords, azCoords]).T
            obs = Observed(altAzs, site=site)
            fc = Observed([[altCen, azCen]], site=site)
            field = Field(obs, field_center=fc)
            obs1 = Observed(field, site=site)
            # with open("/Users/csayres/Desktop/last.txt", "w") as f:
            #     f.write("%s %.8f %.8f\n"%(site.name, azCen, altCen))
            #     for coord in altAzs:
            #         f.write("%.8f %.8f\n"%(coord[0], coord[1]))

            numpy.testing.assert_array_almost_equal(obs, obs1, decimal=9)
            numpy.testing.assert_array_almost_equal(obs.ra, obs1.ra, decimal=9)
            numpy.testing.assert_array_almost_equal(obs.dec, obs1.dec, decimal=9)
            numpy.testing.assert_array_almost_equal(obs.ha, obs1.ha, decimal=9)
            numpy.testing.assert_array_almost_equal(obs.pa, obs1.pa, decimal=9)

            fieldArr = numpy.array(field.copy())
            field1 = Field(fieldArr, field_center=fc)

            numpy.testing.assert_equal(field, field1)
            for attr in ['x', 'y', 'z', 'x_angle', 'y_angle']:
                numpy.testing.assert_almost_equal(
                    getattr(field, attr), getattr(field1, attr), decimal=9
                )


def test_field_to_focal():
    npts = 10
    phis = numpy.zeros(npts) + 2 # two deg off axis
    thetas = numpy.zeros(npts) + 45
    coordArr = numpy.array([thetas, phis]).T
    wls = numpy.random.choice([5400., 6231., 16600.], size=npts)
    site = apoSite
    fc = Observed([[80, 92]], site=site)
    field = Field(coordArr, field_center=fc)
    fp = FocalPlane(field, wavelength=wls, site=site)

    fpArr = numpy.array(fp)
    fp2 = FocalPlane(fpArr, wavelength=wls, site=site)

    numpy.testing.assert_array_equal(fp, fp2)
    numpy.testing.assert_array_equal(fp.b, fp2.b)
    numpy.testing.assert_array_equal(fp.R, fp2.R)


def test_reasonable_field_focal_cycle():
    arcsecError = 0.01 # tenth of an arcsec on cycle

    nCoords = 100000
    for site, maxField in zip(
        [lcoSite, apoSite], [LCO_MAX_FIELD_R, APO_MAX_FIELD_R]
    ):
        fc = Observed([[80, 120]], site=site)
        thetaField = numpy.random.uniform(0, 360, size=nCoords)
        phiField = numpy.random.uniform(0, maxField, size=nCoords)
        wls = numpy.random.choice(list(VALID_WAVELENGTHS), size=nCoords)
        # wls = 16600
        coordArr = numpy.array([thetaField, phiField]).T
        field = Field(coordArr, field_center=fc)
        assert not True in field.field_warn
        fp = FocalPlane(field, wavelength=wls, site=site)
        assert not True in fp.field_warn
        field2 = Field(fp, field_center=fc)
        assert not True in field2.field_warn

        xyz = numpy.array([field.x, field.y, field.z]).T
        xyz2 = numpy.array([field2.x, field2.y, field2.z]).T

        angErrors = []
        for _xyz, _xyz2 in zip(xyz, xyz2):
            # print("norms", numpy.linalg.norm(_xyz), numpy.linalg.norm(_xyz2))
            dt = _xyz.dot(_xyz2)
            if dt > 1:
                assert dt-1 < 1e-8 # make sure it's basically zero
                dt = 1  # nan in arccos if greater than 1
            # print("xyz", _xyz, _xyz2)
            # print("\n\n")
            err = numpy.degrees(numpy.arccos(dt))*3600
            angErrors.append(err)
        angErrors = numpy.array(angErrors)
        maxErr = numpy.max(angErrors)
        assert maxErr < arcsecError


def test_unreasonable_field_focal_cycle():
    # for lco, apo. lco errors remain small despite large field
    arcsecErrs = [1, 11]

    nCoords = 10000
    for site, maxField, errLim in zip(
        [lcoSite, apoSite], [LCO_MAX_FIELD_R, APO_MAX_FIELD_R], arcsecErrs
    ):
        fc = Observed([[80, 120]], site=site)
        thetaField = numpy.random.uniform(0, 360, size=nCoords)
        phiField = numpy.random.uniform(1.5*maxField, 1.6*maxField, size=nCoords)
        wls =  numpy.random.choice(list(VALID_WAVELENGTHS), size=nCoords)
        # wls = 16600.
        coordArr = numpy.array([thetaField, phiField]).T
        field = Field(coordArr, field_center=fc)
        assert not True in field.field_warn
        with pytest.warns(CoordIOUserWarning):
            fp = FocalPlane(field, wavelength=wls, site=site)
        assert not False in fp.field_warn
        with pytest.warns(CoordIOUserWarning):
            field2 = Field(fp, field_center=fc)
        assert not False in field2.field_warn

        xyz = numpy.array([field.x, field.y, field.z]).T
        xyz2 = numpy.array([field2.x, field2.y, field2.z]).T

        angErrors = []
        for _xyz, _xyz2 in zip(xyz, xyz2):
            # print("norms", numpy.linalg.norm(_xyz), numpy.linalg.norm(_xyz2))
            dt = _xyz.dot(_xyz2)
            if dt > 1:
                assert dt-1 < 1e-8 # make sure it's basically zero
                dt = 1  # nan in arccos if greater than 1
            # print("\n\n")
            err = numpy.degrees(numpy.arccos(dt))*3600
            angErrors.append(err)
        angErrors = numpy.array(angErrors)
        maxErr = numpy.max(angErrors)
        print("maxErr", maxErr)
        # break
        assert maxErr < errLim

        # print(angErrors, angErrors.shape)


def test_invalid_wavelength():
    site = apoSite
    nCoords = 100
    fc = Observed([[80, 120]], site=site)
    thetaField = numpy.random.uniform(0, 360, size=nCoords)
    phiField = numpy.random.uniform(0, 1, size=nCoords)
    wls = 250 # invalid
    coordArr = numpy.array([thetaField, phiField]).T
    field = Field(coordArr, field_center=fc)
    with pytest.raises(CoordIOError):
        fp = FocalPlane(field, wavelength=wls, site=site)
    wls = numpy.random.choice([30, 40, 50], size=nCoords)
    with pytest.raises(CoordIOError):
        fp = FocalPlane(field, wavelength=wls, site=site)


def test_invalid_site():
    site = "junk"
    with pytest.raises(CoordIOError):
        fc = Observed([[80, 120]], site=site)
    site = Site("APO") # time not specified
    with pytest.raises(CoordIOError):
        fc = Observed([[80, 120]], site=site)


def test_focal_plane_closest_wavelength():
    site = apoSite
    thetaField = numpy.random.uniform(0, 360, size=1)
    phiField = numpy.random.uniform(0, 1, size=1)
    coordArr = numpy.array([thetaField, phiField]).T

    fc = Observed([[80, 120]], site=site)
    field = Field(coordArr, field_center=fc)
    wls = 100
    fp = FocalPlane(field, wavelength=wls, site=site, use_closest_wavelength=True)

    assert numpy.allclose(fp.wavelength, [5400.0])


def test_focal_plane_closest_wavelength_array():
    site = apoSite
    nCoords = 3
    thetaField = numpy.random.uniform(0, 360, size=nCoords)
    phiField = numpy.random.uniform(0, 1, size=nCoords)
    coordArr = numpy.array([thetaField, phiField]).T

    fc = Observed([[80, 120]], site=site)
    field = Field(coordArr, field_center=fc)
    wls = numpy.array([100, 6500, 8000])
    fp = FocalPlane(field, wavelength=wls, site=site, use_closest_wavelength=True)

    assert numpy.allclose(fp.wavelength, [5400.0, 6231.0, 6231.0])



import ctypes
import importlib
import os

import numpy
import pandas
from scipy.interpolate import interp1d
from scipy import optimize
from functools import partial
from astropy import stats

from . import defaults
from .site import Site
from .sky import ICRS, Observed
from .telescope import Field, FocalPlane
from .wok import Wok

# Get simplexy C function
mod_path = os.path.join(os.path.dirname(__file__), 'libdimage')
success = False
for suffix in importlib.machinery.EXTENSION_SUFFIXES:
    try:
        libPath = mod_path + suffix
        if os.path.exists(libPath):
            success = True
            break
    except OSError:
        pass
if success is False:
    raise OSError('Could not find a valid libdimage extension.')


libdimage = ctypes.cdll.LoadLibrary(libPath)
# simplexy_function = libdimage.simplexy


def wokCurveAPO(r):
    """Curve of the wok at APO at radial position r

    Parameters
    -----------
    r : scalar or 1D array
        radius (cylindrical coords) mm

    Returns:
    ---------
    result : scalar or 1D array
        z (height) of wok surface in mm (0 at vertex)


    Curve that was specified for machining the APO wok, decided by Colby J.
    """
    A = 9199.322517101522
    return A - numpy.sqrt(A**2 - r**2)


def wokCurveLCO(r):
    """Curve of the wok LCO at radial position r

    Parameters
    -----------
    r : scalar or 1D array
        radius (cylindrical coords) mm

    Returns:
    ---------
    result : scalar or 1D array
        z (height) of wok surface in mm (0 at vertex)

    Curve that was specified for machining the LCO wok, decided by Colby J.
    """
    A = 0.000113636363636
    B = 0.0000000129132231405
    C = 0.0000012336318
    return A * r**2 / (1 + numpy.sqrt(1 - B * r**2)) + C * r**2


def radec2wokxy(ra, dec, coordEpoch, waveName, raCen, decCen, obsAngle,
                obsSite, obsTime, focalScale=None, pmra=None, pmdec=None, parallax=None,
                radVel=None, pressure=None, relativeHumidity=0.5,
                temperature=10, fullOutput=False, darLambda=None):
    r"""
    Convert from ra, dec ICRS coords to a flat-wok XY in mm.  At obsAngle=0
    wok +y is a aligned with +dec, and wok +x is aligned with +ra

    Question for José, do we need to enforce a time scale?  I think everything
    defaults to UTC.

    Parameters
    ------------
    ra : numpy.ndarray
        Right ascension in ICRS, degrees
    dec : numpy.ndarray
        Declination in ICRS, degrees
    coordEpoch : float
        A TDB Julian date, the epoch for the input
        coordinates (from the catalog). Defaults to J2000.
    waveName : str, or numpy.ndarray
        Array elements must be "Apogee", "Boss", or "GFA" strings
    raCen : float
        Right ascension of field center, in degrees
    decCen : float
        Declination of field center, in degrees.
    obsAngle : float
        Position angle for observation.  Angle is measured from North
        through East to wok +y. So obsAngle of 45 deg, wok +y points NE.
    obsSite : str
        Either "APO" or "LCO"
    obsTime : float
        TDB Julian date.  The time at which these coordinates will be observed
        with the FPS.
    focalScale : float or None
        Scale factor for conversion between focal and wok coords, found
        via dither analysis.  Defaults to value in defaults.SITE_TO_SCALE
    pmra : numpy.ndarray
        A 1D array with the proper motion in the RA axis for the N targets,
        in milliarcsec/yr. Must be a true angle, i.e, it must include the
        ``cos(dec)`` term.  Defaults to 0.
    pmdec : numpy.ndarray
        A 1D array with the proper motion in the RA axis for the N targets,
        in milliarcsec/yr.  Defaults to 0.
    parallax : numpy.ndarray
        A 1D array with the parallax for the N targets, in milliarcsec.
        Defaults to 0.
    radVel : numpy.ndarray
        A 1D array with the radial velocity in km/s, positive when receding.
        Defaults to 0.
    pressure : float
        The atmospheric pressure at the site, in millibar (same as hPa). If
        not provided the pressure will be calculated using the altitude
        :math:`h` and the approximate expression

        .. math::

            p \sim -p_0 \exp\left( \dfrac{g h M} {T_0 R_0} \right)

        where :math:`p_0` is the pressure at sea level, :math:`M` is the molar
        mass of the air, :math:`R_0` is the universal gas constant, and
        :math:`T_0=288.16\,{\rm K}` is the standard sea-level temperature.
    relativeHumidity : float
        The relative humidity, in the range :math:`0-1`. Defaults to 0.5.
    temperature : float
        The site temperature, in degrees Celsius. Defaults to
        :math:`10^\circ{\rm C}`.
    fullOutput : Bool
        If True all intermediate coords are returned

    Returns
    ---------
    xWok : numpy.ndarray
        x wok coordinate, mm
    yWok : numpy.ndarray
        y wok coordinate, mm
    fieldWarn : numpy.ndarray
        boolean array.  Where True the coordinate converted should be eyed with
        suspicion.  (It came from very far off axis).
    hourAngle : float
        hour angle of field center in degrees
    positionAngle : float
        position angle of field center in degrees

    if fullOutput=True additionally returns:

    xFocal : numpy.ndarray
        x focal coordinate, mm
    yFocal : numpy.ndarray
        y focal coordinate, mm
    thetaField : numpy.ndarray
        azimuthal field coordinate, deg +RA through +Dec
    phiField : numpy.ndarray
        polar field coordinate, deg off axis
    altitude :  numpy.ndarray
        altitude coordinate, deg
    azimuth : numpy.ndarray
        azimuth coordinate N=0, E=90, deg
    altFieldCen : float
        altitude of field center, deg
    azFieldCen : float
        azimuth of field center, deg

    """
    nCoords = len(ra)

    # first grab the correct wavelengths for fibers
    wavelength = numpy.zeros(nCoords)

    if isinstance(waveName, str):
        # same wl for all coords
        wavelength += defaults.INST_TO_WAVE[waveName]
    else:
        assert len(waveName) == nCoords
        for ii, ft in enumerate(waveName):
            wavelength[ii] = defaults.INST_TO_WAVE[ft]

    site = Site(
        obsSite, pressure=pressure,
        temperature=temperature, rh=relativeHumidity
    )
    site.set_time(obsTime)

    # first determine the field center in observed coordinates
    # use the guide wavelength for field center
    # epoch not needed, no propermotions, etc (josé verify?)
    icrsCen = ICRS([[raCen, decCen]])
    obsCen = Observed(icrsCen, site=site, wavelength=defaults.INST_TO_WAVE["GFA"])

    radec = numpy.array([ra, dec]).T

    if pmra is not None:
        pmra[numpy.isnan(pmra)] = 0

    if pmdec is not None:
        pmdec[numpy.isnan(pmdec)] = 0

    if parallax is not None:
        parallax[numpy.isnan(parallax)] = 0

    icrs = ICRS(
        radec, epoch=coordEpoch, pmra=pmra, pmdec=pmdec,
        parallax=parallax, rvel=radVel
    )

    # propogate propermotions, etc
    icrs = icrs.to_epoch(obsTime, site=site)
    if focalScale is None:
        focalScale = defaults.SITE_TO_SCALE[obsSite]

    # check to see if darLambda was supplied, and if so
    # use that wavelength (just for DAR not telescope distortion model)
    darWavelength = wavelength.copy()
    if darLambda is not None:
        if hasattr(darLambda, "__len__"):
            assert len(darLambda) == nCoords
            darWavelength = darLambda
        else:
            darWavelength.fill(darLambda)
    obs = Observed(icrs, site=site, wavelength=darWavelength)
    field = Field(obs, field_center=obsCen)
    focal = FocalPlane(field,
                       wavelength=wavelength,
                       site=site,
                       fpScale=focalScale,
                       use_closest_wavelength=True)
    wok = Wok(focal, site=site, obsAngle=obsAngle)

    output = (
        wok[:, 0], wok[:, 1], focal.field_warn,
        float(obsCen.ha), float(obsCen.pa)
    )

    if fullOutput:
        output += (
            focal[:, 0], focal[:, 1],
            field[:, 0], field[:, 1],
            obs[:, 0], obs[:, 1],
            obsCen[0][0], obsCen[0][1]
        )
    return output


def wokxy2radec(xWok, yWok, waveName, raCen, decCen, obsAngle,
                obsSite, obsTime, focalScale=None, pressure=None, relativeHumidity=0.5,
                temperature=10):
    r"""
    Convert from flat-wok XY (mm) to ra, dec ICRS coords (deg)

    Question for José, do we need to enforce a time scale?  I think everything
    defaults to UTC.

    Parameters
    ------------
    xWok : numpy.ndarray
        X wok coordinates, mm
    yWok : numpy.ndarray
        Y wok coordinates, mm
    waveName : str, or numpy.ndarray
        Array elements must be "Apogee", "Boss", or "GFA" strings
    raCen : float
        Right ascension of field center, in degrees
    decCen : float
        Declination of field center, in degrees.
    obsAngle : float
        Position angle for observation.  Angle is measured from North
        through East to wok +y. So obsAngle of 45 deg, wok +y points NE.
    obsSite : str
        Either "APO" or "LCO"
    obsTime : float
        TDB Julian date.  The time at which these coordinates will be observed
        with the FPS.
    focalScale : float or None
        Scale factor for conversion between focal and wok coords, found
        via dither analysis.  Defaults to value in defaults.SITE_TO_SCALE
    pressure : float
        The atmospheric pressure at the site, in millibar (same as hPa). If
        not provided the pressure will be calculated using the altitude
        :math:`h` and the approximate expression

        .. math::

            p \sim -p_0 \exp\left( \dfrac{g h M} {T_0 R_0} \right)

        where :math:`p_0` is the pressure at sea level, :math:`M` is the molar
        mass of the air, :math:`R_0` is the universal gas constant, and
        :math:`T_0=288.16\,{\rm K}` is the standard sea-level temperature.
    relativeHumidity : float
        The relative humidity, in the range :math:`0-1`. Defaults to 0.5.
    temperature : float
        The site temperature, in degrees Celsius. Defaults to
        :math:`10^\circ{\rm C}`.

    Returns
    ---------
    xWok : numpy.ndarray
        x wok coordinate, mm
    yWok : numpy.ndarray
        y wok coordinate, mm
    fieldWarn : numpy.ndarray
        boolean array.  Where True the coordinate converted should be eyed with
        suspicion.  (It came from very far off axis).
    """
    nCoords = len(xWok)

    # first grab the correct wavelengths for fibers
    wavelength = numpy.zeros(nCoords)

    if isinstance(waveName, str):
        # same wl for all coords
        wavelength += defaults.INST_TO_WAVE[waveName]
    else:
        assert len(waveName) == nCoords
        for ii, ft in enumerate(waveName):
            wavelength[ii] = defaults.INST_TO_WAVE[ft]

    site = Site(
        obsSite, pressure=pressure,
        temperature=temperature, rh=relativeHumidity
    )
    site.set_time(obsTime)

    # first determine the field center in observed coordinates
    # use the guide wavelength for field center
    # epoch not needed, no propermotions, etc (josé verify?)
    icrsCen = ICRS([[raCen, decCen]])
    obsCen = Observed(icrsCen, site=site, wavelength=defaults.INST_TO_WAVE["GFA"])

    # hack in z wok position of 143 mm
    # could do a little better and estimate the curve of focal surface
    xyzWok = numpy.zeros((nCoords, 3))
    rWok = numpy.sqrt(xWok**2 + yWok**2)

    # this is not totally correct
    # but probably better than modeling as flat
    # project the xy positions onto the 3d surface
    # and add 143 (the positioner height)
    if obsSite == "APO":
        zWok = defaults.POSITIONER_HEIGHT + wokCurveAPO(rWok)
    else:
        zWok = defaults.POSITIONER_HEIGHT + wokCurveLCO(rWok)

    xyzWok[:, 0] = xWok
    xyzWok[:, 1] = yWok
    xyzWok[:, 2] = zWok

    wok = Wok(xyzWok, site=site, obsAngle=obsAngle)
    if focalScale is None:
        focalScale = defaults.SITE_TO_SCALE[obsSite]
    focal = FocalPlane(wok,
                       wavelength=wavelength,
                       site=site,
                       fpScale=focalScale,
                       use_closest_wavelength=True)
    field = Field(focal, field_center=obsCen)
    obs = Observed(field, site=site, wavelength=wavelength)
    icrs = ICRS(obs, epoch=obsTime)

    return icrs[:, 0], icrs[:, 1], field.field_warn


def fitsTableToPandas(recarray):
    d = {}
    if hasattr(recarray, "names"):
        names = recarray.names
    else:
        names = recarray.dtype.names
    for name in names:
        key = name
        if hasattr(key, "decode"):
            key = key.decode()
        d[key] = recarray[name].byteswap().newbyteorder()

    df = pandas.DataFrame(d)
    # decode binary data into strings, if present
    # https://stackoverflow.com/questions/40389764/how-to-translate-bytes-objects-into-literal-strings-in-pandas-dataframe-pytho
    # str_df = df.select_dtypes([numpy.object])
    # if len(str_df) > 0:
    #     str_df = str_df.stack().str.decode('utf-8').unstack()
    #     for col in str_df:
    #         df[col] = str_df[col]
    return df


def simplexy(image, psf_sigma=1., plim=8., dlim=1., saddle=3., maxper=1000,
             maxnpeaks=5000):
    """Determines positions of stars in an image.
    Parameters
    ----------
    image : numpy.float32
        2-D ndarray
    psf_sigma : float
        sigma of Gaussian PSF to assume (default 1 pixel)
    plim : float
        significance to select objects on (default 8)
    dlim : float
        tolerance for closeness of pairs of objects (default 1 pixel)
    saddle : float
        tolerance for depth of saddle point to separate sources
        (default 3 sigma)
    maxper : int
        maximum number of children per parent (default 1000)
    maxnpeaks : int
        maximum number of stars to find total (default 100000)
    Returns
    -------
    (x, y, flux) : (numpy.float32, numpy.float32, numpy.float32)
         ndarrays with pixel positions and peak pixel values of stars
    Notes
    -----
    Calls simplexy.c in libdimage.so
    copied directly from: https://github.com/blanton144/dimage
    """

    # Create image pointer
    if(image.dtype != numpy.float32):
        image_float32 = image.astype(numpy.float32)
        image_ptr = image_float32.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    else:
        image_ptr = image.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

    nx = image.shape[0]
    ny = image.shape[1]
    psf_sigma_ptr = ctypes.c_float(psf_sigma)
    plim_ptr = ctypes.c_float(plim)
    dlim_ptr = ctypes.c_float(dlim)
    saddle_ptr = ctypes.c_float(saddle)
    maxper_ptr = ctypes.c_int(maxper)
    maxnpeaks_ptr = ctypes.c_int(maxnpeaks)

    x = numpy.zeros(maxnpeaks, dtype=numpy.float32)
    x_ptr = x.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    y = numpy.zeros(maxnpeaks, dtype=numpy.float32)
    y_ptr = y.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    flux = numpy.zeros(maxnpeaks, dtype=numpy.float32)
    flux_ptr = flux.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    sigma = ctypes.c_float(0.)
    npeaks = ctypes.c_int(0)

    libdimage.simplexy(
        image_ptr, nx, ny, psf_sigma_ptr, plim_ptr,
        dlim_ptr, saddle_ptr, maxper_ptr, maxnpeaks_ptr,
        ctypes.byref(sigma), x_ptr, y_ptr, flux_ptr,
        ctypes.byref(npeaks)
    )

    npeaks = npeaks.value
    x = x[0:npeaks]
    y = y[0:npeaks]
    flux = flux[0:npeaks]

    return (x, y, flux)


def refinexy(image, x, y, psf_sigma=2., cutout=19):
    """Refines positions of stars in an image.
    Parameters
    ----------
    image : numpy.float32
        2-D ndarray
    x : numpy.float32
        1-D ndarray of rough x positions
    y : numpy.float32
        1-D ndarray of rough y positions
    psf_sigma : float
        sigma of Gaussian PSF to assume (default 2 pixels)
    cutout : int
        size of cutout used, should be odd (default 19)
    Returns
    -------
    xr : ndarray of numpy.float32
        refined x positions
    yr : ndarray of numpy.float32
        refined y positions
    Notes
    -----
    Calls drefine.c in libdimage.so
    copied directly from: https://github.com/blanton144/dimage
    """

    # Create image pointer
    if(image.dtype != numpy.float32):
        image_float32 = image.astype(numpy.float32)
        image_ptr = image_float32.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    else:
        image_ptr = image.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

    nx = image.shape[0]
    ny = image.shape[1]
    psf_sigma_ptr = ctypes.c_float(psf_sigma)

    ncen = len(x)
    ncen_ptr = ctypes.c_int(ncen)
    cutout_ptr = ctypes.c_int(cutout)
    xrough = numpy.float32(x)
    xrough_ptr = xrough.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    yrough = numpy.float32(y)
    yrough_ptr = yrough.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

    xrefined = numpy.zeros(ncen, dtype=numpy.float32)
    xrefined_ptr = xrefined.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
    yrefined = numpy.zeros(ncen, dtype=numpy.float32)
    yrefined_ptr = yrefined.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

    libdimage.drefine(
        image_ptr, nx, ny,
        xrough_ptr, yrough_ptr, xrefined_ptr, yrefined_ptr,
        ncen_ptr, cutout_ptr, psf_sigma_ptr
    )

    return (xrefined, yrefined)


class MoffatLossProfile(object):
    """
    class for calculating fiber loss based on Moffat profile

    Parameters
    ----------
    offset: numpy.array
        fiber offset

    beta: float
        Power index of the Moffat profile

    FWHM: float
        seeing

    rfiber: float
        radius of the fiber
    """
    def __init__(self, offset, beta, FWHM, rfiber=1):
        self.offset = offset
        self.beta = beta
        self.FWHM = FWHM
        self.alpha = self.FWHM / (2. * numpy.sqrt(2 ** (1 / self.beta) - 1))
        self.amplitude = 1. / self.moffat_norm(1.)  # this is the amplitude needed to normalize the Moffat profile
        self.rfiber = rfiber

    def moffat_1D(self, r, amplitude=None):  # unit flux at the center
        """
        Function computing Moffat 1D profile

        Parameters
        ------------
        r: float or numpy.array
            Distance from the centre of the profile

        Returns
        -------
        moff_prof: float or numpy.array
            1D Moffat profile.
        """
        if amplitude is None:
            moff_prof = self.amplitude * (1 + (r / self.alpha) ** 2) ** (-self.beta)
        else:
            moff_prof = amplitude * (1 + (r / self.alpha) ** 2) ** (-self.beta)
        return moff_prof

    def moffat_norm(self, amplitude):
        """
        Function computing the normalized the Moffat 1D profile.

        Parameters
        -----------
        amplitude: float
            Amplitude of the profile

        Returns
        -------
        moff_prof_norm: float
            Normalised Moffat profile.
        """

        xmin, xmax, xstep = -7., 7., 0.05  # fixed ln radius steps
        # x is the ln radius in units of the r / alpha
        steps = numpy.arange(xmin, xmax, xstep)
        r = self.alpha * numpy.exp(steps)
        norm = numpy.sum(numpy.exp(2 * steps) * self.moffat_1D(r, amplitude=amplitude))
        moff_prof_norm = 2 * numpy.pi * self.alpha ** 2 * norm * xstep
        return moff_prof_norm

    def flux_loss(self, offset):
        """
        Function computing the flux loss obtained by moving the fiber
        across the source in a certain direction.

        Prameters
        ---------
        offset: numpy.array
            fiber offset

        Returns
        --------
        norm: numpy.array
            the flux loss
        """
        x = numpy.arange(-self.rfiber, self.rfiber, self.rfiber / 50)    # set up a 100x100 Cartesian grid across the fiber
        y = numpy.arange(-self.rfiber, self.rfiber, self.rfiber / 50)
        X, Y = numpy.meshgrid(x, y)
        r = numpy.sqrt(X **2 + Y ** 2)
        norm = numpy.zeros(len(offset))
        r_ev = (r <= self.rfiber)
        for i in range(len(offset)):
            r_moffat = numpy.sqrt((X - offset[i]) ** 2 + Y ** 2)
            norm[i] = numpy.sum((self.rfiber / 50) ** 2 *
                                self.moffat_1D(r_moffat[r_ev]))
        return norm

    def func_magloss(self):
        """
        Function computing the magnitude loss obtained by moving the fiber
        across the source in a certain direction. Note that the magnitude
        loss with offset = 0 is not 0, and should coincide with the difference
        between PSF and fiber magnitudes.

        Prameters
        ----------

        Returns
        --------
        magloss: numpy.array
            the magnitude loss
        """
        magloss = numpy.zeros(len(self.offset))
        magloss = -2.5 * numpy.log10(self.flux_loss(self.offset)) \
                   + 2.5 * numpy.log10(self.flux_loss(numpy.array([0.]))[0])
        return magloss


def default_Moffat_params():
    """
    Returns the dict of default parameters for
    the Moffat profile base don observator and lunation
    """
    # set beta
    beta = {}
    beta['bright'] = {'APO': 1.6, 'LCO': 1.7}
    beta['dark'] = {'APO': 1.9, 'LCO': 1.8}
    # set FWHM
    FWHM = {}
    FWHM['bright'] = {'APO': 0.5, 'LCO': 0.8}
    FWHM['dark'] = {'APO': 1.4, 'LCO': 1.1}
    return beta, FWHM


class Moffat2dInterp(object):
    """
    Create the dict of 1D interpolations function
    for a moffat profile offset for various FWHMs.
    Object returned includes two dicts, one for APO
    and one for LCO for the various fiber sizes.
    """
    def __init__(self, Noffset=None, FWHM=None, beta=None):
        if Noffset is None:
            Noffset = 1500
        if FWHM is None:
            defualt = default_Moffat_params()[1]
            FWHM = [defualt[l][o] for l in ['bright', 'dark'] for o in ['APO', 'LCO']]
        if beta is None:
            beta = default_Moffat_params()[0]
        rfibers = {'APO': 1., 'LCO': 1.33 / 2}
        offsets = numpy.zeros((len(FWHM), Noffset))
        FWHMs = numpy.zeros((len(FWHM), Noffset))
        for i, f in enumerate(FWHM):
            FWHMs[i, :] = f
            offsets[i, :] = numpy.linspace(0, 50, Noffset)

        magloss = numpy.zeros((FWHMs.shape[0], Noffset))

        fmagloss = {}
        # this is for default case to have it by lunation as well
        if isinstance(beta, dict) and 'bright' in beta.keys():
            for lunation, beta_lun in zip(beta.keys(), beta.values()):
                fmagloss[lunation] = {}
                for obs, rfiber in zip(rfibers.keys(), rfibers.values()):
                    fmagloss[lunation][obs] = {}
                    b = beta_lun[obs]
                    for i, f in enumerate(FWHMs[:, 0]):
                        magloss[i, :] = MoffatLossProfile(offsets[i, :], b, f, rfiber=rfiber).func_magloss()
                        fmagloss[lunation][obs][f] = interp1d(magloss[i, :], offsets[i, :])
        else:
            for obs, rfiber in zip(rfibers.keys(), rfibers.values()):
                fmagloss[obs] = {}
                if isinstance(beta, dict):
                    b = beta[obs]
                else:
                    b = beta
                for i, f in enumerate(FWHMs[:, 0]):
                    magloss[i, :] = MoffatLossProfile(offsets[i, :], b, f, rfiber=rfiber).func_magloss()
                    fmagloss[obs][f] = interp1d(magloss[i, :], offsets[i, :])
        self.fmagloss = fmagloss
        self.beta_interp2d = beta
        self.FWHM_interp2d = FWHM

    def __call__(self, magloss, FWHM, obsSite, lunation=None):
        """
        The cal to return the offset value based on the desired
        magloss.

        Parameters
        -----------
        magloss: float or numpy.array
            The desired magnitude loss for the object(s)

        FWHM: float
            The FWHM for the moffat profile that has been calculated
            on the init of the object

        obsSite: str
            The observatory of the observation. Should either be
            'APO' or 'LCO'.

        lunation: str
            If the designmode is bright time ('bright') or dark
            time ('dark'). Required if set up using default params.

        Returns
        -------
        r: float or numpy.array
            The offset to get the desired magloss in arcseconds.
        """
        if 'bright' in self.fmagloss.keys():
            if lunation is None:
                raise ValueError('Must provide lunation for default function!')
            r = self.fmagloss[lunation][obsSite][FWHM](magloss)
        else:
            r = self.fmagloss[obsSite][FWHM](magloss)
        return r


def offset_definition(mag, mag_limits, lunation, waveName, obsSite, fmagloss=None,
                      safety_factor=0., beta=None, FWHM=None, skybrightness=None,
                      offset_min_skybrightness=None, can_offset=None,
                      use_type='bright_neigh', mag_limit_ind=None):
    """
    Returns the offset needed for object with mag to be
    observed at mag_limit.
    This is for the piecewise appromixation used based on:
    https://wiki.sdss.org/pages/viewpage.action?pageId=100173069

    Parameters
    ----------
    mag: float or numpy.array
        The magniutde(s) of the objects. For BOSS should be
        Gaia G-band and for APOGEE should be 2MASS H-band.

    mag_limits: numpy.array
        Magnitude limits for the designmode of the design.
        This should be an array of length N=10 where indexes
        correspond to magntidues: [g, r, i, z, bp, gaia_g, rp, J, H, K].
        This matches the apogee_bright_limit_targets_min or
        boss_bright_limit_targets_min (depending on instrument) from
        targetdb.DesignMode for the design_mode of the design.

    lunation: str:
        If the designmode is bright time ('bright') or dark
        time ('dark')

    waveName: str
        Instrument for the fibers offset definition being applied
        to. Either 'Boss' or 'Apogee'.

    obsSite: str
        The observatory of the observation. Should either be
        'APO' or 'LCO'.

    fmagloss: object
        Moffat2dInterp class with the lookup table
        for doing the moffat profile inversion. If None,
        then table is calculated at function call.

    safety_factor: float
        Factor to add to mag_limit. Should equal zero for
        bright neighbor checks (i.e. remain at default).

    beta: float
        Power index of the Moffat profile. If None, default set in code.

    FWHM: float
        seeing for the Moffat profile. If None, default set in code.

    skybrightness: float
        Sky brightness for the field cadence. Only set if
        want to check for offset_flag TOO_DARK (8).

    offset_min_skybrightness: float
        Minimum sky brightness for the offset. Only set if
        want to check for offset_flag TOO_DARK (8).

    can_offset: boolean or numpy.array
        can_offset value from targetdb for the target(s) to be
        offset. Only set if
        want to check for offset_flag NO_CAN_OFFSET (16).

    use_type: str
        Defines type for definition use. 'bright_neigh' is for bright
        neighbor check and auto sets mag_limit from mag_limits array.
        'offfset' is for offsetting and the index from mag_limits array
        is set by mag_limit_ind.

    mag_limit_ind: int
        when used with use_type='offfset', then this sets what index
        from mag_limits array is set as the magnitude limit. Otherwise,
        for bright neighbor check this is set in code.

    Returns
    -------
    r: float or numpy.array
        offset radius in arcseconds around object(s)

    offset_flag: int or numpy.array
        bitmask for how offset was set. Flags are:
            - 0: offset applied normally (i.e. when mag <= mag_limit)
            - 1: no offset applied because mag > mag_limit
            - 2: no offset applied because magnitude was null value.
            - 8: No offset applied because sky brightness is <=
                 minimum offset sky brightness
            - 16: no offsets applied because can_offset = False
            - 32: no offset applied because mag <= offset_bright_limit
                  (offset_bright_limit is G = 6 for Boss bright time and
                   G = 13 for Boss dark time, and
                   H = 1 for Apogee).
    """
    # define Null cases for targetdb.magnitude table
    cases = [-999, -9999, 999,
             0.0, numpy.nan, 99.9, None]
    # set magntiude limit for instrument and lunation
    if use_type == 'bright_neigh':
        if waveName == 'Apogee':
            # 2MASS H
            mag_limit = mag_limits[8]
        elif lunation == 'bright':
            # Gaia G
            mag_limit = mag_limits[5]
        else:
            # SDSS r
            mag_limit = mag_limits[1]

        # for bright_neigh, need exclusion radius for all stars
        offset_bright_limit = -9999.
    else:
        mag_limit = mag_limits[mag_limit_ind]

        # only set real mag_limits for offsets
        if waveName == 'Boss':
            if lunation == 'bright':
                offset_bright_limit = 6.
            else:
                offset_bright_limit = 13.
        else:
            offset_bright_limit = 1.
    # assign correct FWHM
    if FWHM is None:
        FWHM = default_Moffat_params()[1][lunation][obsSite]
    if beta is None:
        beta = default_Moffat_params()[0]
    # get magloss function
    if fmagloss is None:
        fmagloss = Moffat2dInterp(beta=beta, FWHM=[FWHM])
    if isinstance(mag, float) or isinstance(mag, int):
        # make can_offset always True if not supplied
        if can_offset is None:
            can_offset = True
        offset_flag = 0
        if mag <= mag_limit and mag not in cases and can_offset and mag > offset_bright_limit:
            # linear portion in the wings
            r_wings = ((mag_limit + safety_factor) - mag - 8.2) / 0.05
            # core area
            if beta != fmagloss.beta_interp2d or FWHM not in fmagloss.FWHM_interp2d:
                fmagloss = Moffat2dInterp(beta=beta, FWHM=[FWHM])
                r_core = fmagloss((mag_limit + safety_factor) - mag, FWHM, obsSite, lunation=lunation)
            else:
                r_core = fmagloss((mag_limit + safety_factor) - mag, FWHM, obsSite, lunation=lunation)
            # tom's old conservative core function
            # r_core = 1.5 * ((mag_limit + safety_factor) - mag) ** 0.8
            # exlusion radius is the max of each section
            r = numpy.nanmax([r_wings, r_core])
        else:
            r = 0.
            if mag > mag_limit:
                offset_flag += 1
            elif mag in cases:
                offset_flag += 2
            elif not can_offset:
                offset_flag += 16
            else:
                offset_flag += 32
        if skybrightness is not None and offset_min_skybrightness is not None:
            if skybrightness <= offset_min_skybrightness:
                offset_flag += 8
                r = 0.
    else:
        # make can_offset always True if not supplied
        if can_offset is None:
            can_offset = numpy.zeros(mag.shape, dtype=bool) + True
        # create empty arrays for each portion
        r_wings = numpy.zeros(mag.shape)
        r_core = numpy.zeros(mag.shape)
        # only do calc for valid mags and can_offsets for offset
        # to avoid warning
        mag_valid = ((mag <= mag_limit) & (~numpy.isin(mag, cases)) &
                     (~numpy.isnan(mag)) & (can_offset) & (mag > offset_bright_limit))
        # set flags
        offset_flag = numpy.zeros(mag.shape, dtype=int)
        offset_flag[mag > mag_limit] += 1
        offset_flag[(numpy.isin(mag, cases)) | (numpy.isnan(mag))] += 2
        offset_flag[~can_offset] += 16
        offset_flag[(mag <= offset_bright_limit) & ~((numpy.isin(mag, cases)) | (numpy.isnan(mag)))] += 32
        # linear portion in the wings
        r_wings[mag_valid] = ((mag_limit + safety_factor) - mag[mag_valid] - 8.2) / 0.05
        # core area
        if beta != fmagloss.beta_interp2d or FWHM not in fmagloss.FWHM_interp2d:
            fmagloss = Moffat2dInterp(beta=beta, FWHM=[FWHM])
            r_core[mag_valid] = fmagloss((mag_limit + safety_factor) - mag[mag_valid],
                                         FWHM, obsSite, lunation=lunation)
        else:
            r_core[mag_valid] = fmagloss((mag_limit + safety_factor) - mag[mag_valid],
                                         FWHM, obsSite, lunation=lunation)
        # tom's old conservative core function
        # r_core[mag_valid] = 1.5 * ((mag_limit + safety_factor) - mag[mag_valid]) ** 0.8
        # exlusion radius is the max of each section
        r = numpy.nanmax(numpy.column_stack((r_wings,
                                             r_core)),
                         axis=1)
        if skybrightness is not None and offset_min_skybrightness is not None:
            if skybrightness <= offset_min_skybrightness:
                offset_flag[:] += 8
                r[:] = 0.
    return r, offset_flag


def object_offset(mags, mag_limits, lunation, waveName, obsSite, fmagloss=None,
                  safety_factor=None, beta=None, FWHM=None, skybrightness=None,
                  offset_min_skybrightness=None, can_offset=None):
    """
    Returns the offset needed for object with mag to be
    observed at mag_limit. Currently assumption is all offsets
    are set in positive RA direction.

    Parameters
    ----------
    mags: numpy.array
        The magniutdes of the objects. Should be a 2D, Nx10 array, where
        N is number of objects and length of 10 index should correspond
        to magntidues: [g, r, i, z, bp, gaia_g, rp, J, H, K].

    mag_limits: numpy.array
        Magnitude limits for the designmode of the design.
        This should be an array of length N=10 where indexes
        correspond to magntidues: [g, r, i, z, bp, gaia_g, rp, J, H, K].
        This matches the apogee_bright_limit_targets_min or
        boss_bright_limit_targets_min (depending on instrument) from
        targetdb.DesignMode for the design_mode of the design.

    lunation: str:
        If the designmode is bright time ('bright') or dark
        time ('dark')

    waveName: str
        Instrument for the fibers offset definition being applied
        to. Either 'Boss' or 'Apogee'.

    obsSite: str
        The observatory of the observation. Should either be
        'APO' or 'LCO'.

    fmagloss: object
        Moffat2dInterp class with the lookup table
        for doing the moffat profile inversion. If None,
        then table is calculated at function call.

    safety_factor: float
        Factor to add to mag_limit. If None, default set in code.

    beta: float
        Power index of the Moffat profile. If None, default set in code.

    FWHM: float
        seeing for the Moffat profile. If None, default set in code.

    skybrightness: float
        Sky brightness for the field cadence. Only set if
        want to check for offset_flag TOO_DARK (8).

    offset_min_skybrightness: float
        Minimum sky brightness for the offset. Only set if
        want to check for offset_flag TOO_DARK (8).

    can_offset: boolean or numpy.array
        can_offset value from targetdb for the target(s) to be
        offset. Only set if
        want to check for offset_flag NO_CAN_OFFSET (16).

    Returns
    -------
    delta_ra: numpy.array
        offset in RA in arcseconds around object(s) in
        numpy array of length N

    delta_dec: numpy.array
        offset in Decl. in arcseconds around object(s) in
        numpy array of length N

    offset_flag: numpy.array
        bitmask for how offset was set in
        numpy array of length N. Flags are:
            - 0: offset applied normally (i.e. when mag <= mag_limit)
            - 1: no offset applied because mag > mag_limit
            - 2: no offset applied because magnitude was null value.
            - 8: offsets should not be used as sky brightness is <=
                 minimum offset sky brightness
            - 16: no offsets applied because can_offset = False
            - 32: no offset applied because mag <= offset_bright_limit
                  (offset_bright_limit is G = 6 for Boss bright time and
                   G = 13 for Boss dark time, and
                   H = 1 for Apogee).
            - 64: no offset applied because no valid magnitude limits
    """
    # check if 2D array
    if len(mags.shape) != 2:
        raise ValueError('mags must be a 2D numpy.array of shape (N, 10)')
    if mags.shape[1] != 10:
        raise ValueError('mags must be a 2D numpy.array of shape (N, 10)')
    # add default values for offset function if None supplied
    if safety_factor is None:
        if lunation == 'bright':
            safety_factor = 0.5
        else:
            safety_factor = 1.0
    if FWHM is None:
        FWHM = default_Moffat_params()[1][lunation][obsSite]
    if beta is None:
        beta = default_Moffat_params()[0]
    delta_ras = numpy.zeros(mags.shape)
    offset_flags = numpy.zeros(mags.shape)
    for i in range(len(mag_limits)):
        if mag_limits[i] != -999.:
            delta_ras[:, i], offset_flags[:, i] = offset_definition(mags[:, i], mag_limits, lunation, waveName,
                                                                    obsSite, fmagloss=fmagloss,
                                                                    safety_factor=safety_factor, beta=beta,
                                                                    FWHM=FWHM, skybrightness=skybrightness,
                                                                    offset_min_skybrightness=offset_min_skybrightness,
                                                                    can_offset=can_offset,
                                                                    use_type='offset', mag_limit_ind=i)
        else:
            # make artificially less than 0 so this doesnt get chosen
            # for max when setting offset_flag
            delta_ras[:, i] = numpy.zeros(len(mags[:, i]))
            offset_flags[:, i] = numpy.zeros(len(mags[:, i])) + 64
    # retain all flags when delta_ra = 0 for all checks
    def unique_offset_flags(flags):
        unq_flags = []
        poss_flags = [1, 2, 8, 16, 32, 64]
        for f in flags:
            for pf in poss_flags:
                if pf & int(f):
                    unq_flags.append(pf)
        total_flags = numpy.sum(numpy.unique(unq_flags))
        return numpy.zeros(flags.shape) + total_flags
    try:
        offset_flags[numpy.all(delta_ras == 0., 1)] = numpy.apply_along_axis(unique_offset_flags,
                                                                             1,
                                                                             offset_flags[numpy.all(delta_ras == 0., 1)])
    except ValueError:
        pass
    # use max offset
    delta_ra = numpy.max(delta_ras, axis=1)
    ind_max = numpy.argmax(delta_ras, axis=1)
    offset_flag = numpy.array([offset_flags[i, j] for i, j in enumerate(ind_max)],
                              dtype=int)
    delta_dec = numpy.zeros(len(delta_ra))
    return delta_ra, delta_dec, offset_flag

def _offset_radec(ra=None, dec=None, delta_ra=0., delta_dec=0.):
    """Offsets ra and dec according to specified amount. From Mike's
    robostrategy.Field object

    Parameters
    ----------
    ra : numpy.float64 or ndarray of numpy.float64
    right ascension, deg
    dec : numpy.float64 or ndarray of numpy.float64
        declination, deg
    delta_ra : numpy.float64 or ndarray of numpy.float64
        right ascension direction offset, arcsec
    delta_dec : numpy.float64 or ndarray of numpy.float64
        declination direction offset, arcsec

    Returns
    -------
    offset_ra : numpy.float64 or ndarray of numpy.float64
        offset right ascension, deg
    offset_dec : numpy.float64 or ndarray of numpy.float64
        offset declination, deg

    Notes
    -----
    Assumes that delta_ra, delta_dec are in proper coordinates; i.e.
    an offset of delta_ra=1 arcsec represents the same angular separation
    on the sky at any declination.
    Carefully offsets in the local directions of ra, dec based on
    the local tangent plane (i.e. does not just scale delta_ra by
    1/cos(dec))
    """
    deg2rad = numpy.pi / 180.
    arcsec2rad = numpy.pi / 180. / 3600.
    x = numpy.cos(dec * deg2rad) * numpy.cos(ra * deg2rad)
    y = numpy.cos(dec * deg2rad) * numpy.sin(ra * deg2rad)
    z = numpy.sin(dec * deg2rad)
    ra_x = - numpy.sin(ra * deg2rad)
    ra_y = numpy.cos(ra * deg2rad)
    ra_z = 0.
    dec_x = - numpy.sin(dec * deg2rad) * numpy.cos(ra * deg2rad)
    dec_y = - numpy.sin(dec * deg2rad) * numpy.sin(ra * deg2rad)
    dec_z = numpy.cos(dec * deg2rad)
    xoff = x + (ra_x * delta_ra + dec_x * delta_dec) * arcsec2rad
    yoff = y + (ra_y * delta_ra + dec_y * delta_dec) * arcsec2rad
    zoff = z + (ra_z * delta_ra + dec_z * delta_dec) * arcsec2rad
    offnorm = numpy.sqrt(xoff**2 + yoff**2 + zoff**2)
    xoff = xoff / offnorm
    yoff = yoff / offnorm
    zoff = zoff / offnorm
    decoff = numpy.arcsin(zoff) / deg2rad
    raoff = ((numpy.arctan2(yoff, xoff) / deg2rad) + 360.) % 360.
    return(raoff, decoff)


def gaia_mags2sdss_gri(gaia_g, gaia_bp=None, gaia_rp=None, gaia_bp_rp=None):
    '''
    Stolen from Tom Dwelly's dither_work:

    https://github.com/sdss/dither_work/blob/main/python/dither_work/utils.py#L75

    There are some fairly accurate+reliable G,BP-RP -> SDSS gri
    transforms available in the literature. Evans et al (2018,
    https://www.aanda.org/articles/aa/full_html/2018/08/aa32756-18/aa32756-18.html)
    give transforms for main sequence stars. Paraphrased below:

    sdss_g_from_gdr2 = phot_g_mean_mag_gdr2 - (0.13518 - 0.46245*bp_rp_gdr2 +
                                               -0.25171*bp_rp_gdr2**2 + 0.021349*bp_rp_gdr2**3)

    sdss_r_from_gdr2 = phot_g_mean_mag_gdr2 - (-0.12879 + 0.24662*bp_rp_gdr2 +
                                               -0.027464*bp_rp_gdr2**2 - 0.049465*bp_rp_gdr2**3)

    sdss_i_from_gdr2 = phot_g_mean_mag_gdr2 - (-0.29676 + 0.64728*bp_rp_gdr2 +
                                               -0.10141*bp_rp_gdr2**2)

    '''
    if (gaia_bp is not None) and (gaia_rp is not None) and (gaia_bp_rp is None):
        gaia_bp_rp = gaia_bp - gaia_rp
    elif (gaia_bp is None) and (gaia_rp is None) and (gaia_bp_rp is not None):
        pass
    else:
        raise Exception("Error - you must supply either bp and rp or just bp-rp")

    sdss_g = gaia_g - (0.13518 - 0.46245 * gaia_bp_rp -
                       0.25171 * gaia_bp_rp**2 + 0.021349 * gaia_bp_rp**3)
    sdss_r = gaia_g - (-0.12879 + 0.24662 * gaia_bp_rp -
                       0.027464 * gaia_bp_rp**2 - 0.049465 * gaia_bp_rp**3)
    sdss_i = gaia_g - (-0.29676 + 0.64728 * gaia_bp_rp -
                       0.10141 * gaia_bp_rp**2)

    return sdss_g, sdss_r, sdss_i


def calc_R(xc, yc, x, y):
    """ calculate the distance of each 2D points from the center (xc, yc) """
    return numpy.sqrt((x-xc)**2 + (y-yc)**2)


def f_2(c, x, y):
    """ calculate the algebraic distance between the data points and the mean circle centered at c=(xc, yc) """
    Ri = calc_R(c[0], c[1], x, y)
    return Ri - Ri.mean()


def fit_circle(xs, ys, sigma_clip=None):
    """
    Fit  a circle to xy inputs xs, ys

    If sigma clip, iteratavely throw out xs ys outliers until no
    outliers remain.

    Based on https://scipy-cookbook.readthedocs.io/items/Least_Squares_Circle.html

    returns:
        xcenter: x center of the circle
        ycenter: y center of the circle
        radius: radius of the circle
        keep: ndarray indicating indices in xs/ys used for fitting.

    """
    xs = numpy.array(xs)
    ys = numpy.array(ys)

    center_estimate = [xs.mean(), ys.mean()]
    keep = numpy.array([True]*len(xs))

    xkeep = xs[keep]
    ykeep = ys[keep]
    _f_2 = partial(f_2, x=xkeep, y=ykeep)
    center_fit, ier = optimize.leastsq(_f_2, center_estimate)
    rs = calc_R(center_fit[0], center_fit[1], xkeep, ykeep)
    radius_fit = rs.mean()

    if sigma_clip is not None:
        # throw out data until there are no outliers
        mask = stats.sigma_clip(rs, sigma=sigma_clip, masked=True)
        keep = ~mask.mask
        xkeep = xs[keep]
        ykeep = ys[keep]
        _f_2 = partial(f_2, x=xkeep, y=ykeep)
        center_fit, ier = optimize.leastsq(_f_2, center_estimate)
        rs = calc_R(center_fit[0], center_fit[1], xkeep, ykeep)
        radius_fit = rs.mean()

    return center_fit[0], center_fit[1], radius_fit, keep











import numpy
import pytest


numpy.random.seed(7)
import time

import matplotlib.pyplot as plt
import pandas
import seaborn as sns

from coordio import conv, libcoordio
# from coordio import Site, Wok, Observed, Field, FocalPlane, Tangent
from coordio.defaults import MICRONS_PER_MM, POSITIONER_HEIGHT, calibration


# import matplotlib.pyplot as plt
# import seaborn as sns


# LEFTHAND = True



# get tangent coords at these locations
holeIDs = ["R0C1", "R+13C1", "R+13C14", "R0C27", "R-13C14", "R+13C7", "R-13C1"]

tanCoordList = [
    [22, -4, 0],
    [-4, 180, 2],
    [60, 4, -1]
]

angTol = 1e-5 ## Deg, much less than a fraction of a micron for a positioner

modelMetXY = numpy.array([14.314, 0])
# boss xy position in solid model
modelBossXY = numpy.array([14.965, -0.376])
# apogee xy position in solid model
modelApXY = numpy.array([14.965, 0.376])
la = 7.4


def test_tangentAndPositionerLib():
    alpha = 0
    beta = 0
    xb = 14
    yb = 0
    alphaLen = 7 # equal arm lengths
    alphaOff = 0
    betaOff = 0
    # tz = 0

    tx, ty = libcoordio.positionerToTangent([alpha, beta], [xb,yb], alphaLen, alphaOff, betaOff)

    assert tx == pytest.approx(alphaLen + xb)
    assert ty == pytest.approx(0)

    _alpha, _beta = libcoordio.tangentToPositioner(
        [tx,ty], [xb,yb], alphaLen, alphaOff, betaOff
    )


    assert _alpha == pytest.approx(alpha, rel=angTol, abs=angTol)
    assert _beta == pytest.approx(beta, rel=angTol, abs=angTol)


    betaOff = 5

    tx, ty = libcoordio.positionerToTangent([alpha,beta], [xb,yb], alphaLen, alphaOff, betaOff)

    _alpha, _beta = libcoordio.tangentToPositioner(
        [tx,ty], [xb,yb], alphaLen, alphaOff, betaOff
    )

    assert (_alpha == pytest.approx(alpha) or _alpha == pytest.approx(alpha+360))
    assert _beta == pytest.approx(beta, rel=angTol, abs=angTol)
    assert ty > 1


    betaOff = -5 # its gotta go right handed
    tx, ty = libcoordio.positionerToTangent([alpha,beta], [xb,yb], alphaLen, alphaOff, betaOff)


    _alpha, _beta = libcoordio.tangentToPositioner(
        [tx,ty], [xb,yb], alphaLen, alphaOff, betaOff
    )

    assert _beta == pytest.approx(-2*betaOff)

    betaOff = 0
    alphaOff = 5

    tx, ty = libcoordio.positionerToTangent([alpha,beta], [xb,yb], alphaLen, alphaOff, betaOff)
    assert ty > 1


def test_tangentAndPositionerSafeSingle():

    alphas = numpy.random.uniform(10,350)
    betas = numpy.random.uniform(10, 170)
    yBetas = numpy.random.uniform(-0.5,0.5)
    xBetas = 15
    la = 7.4
    alphaOff = numpy.random.uniform(-1,1)
    betaOff = numpy.random.uniform(-1,1)

    xt1, yt1 = conv._positionerToTangent(
        alphas, betas, xBetas, yBetas, la, alphaOff, betaOff
    )

    xt2, yt2 = conv.positionerToTangent(
        alphas, betas, xBetas, yBetas, la, alphaOff, betaOff
    )


    assert xt1 == pytest.approx(xt2)
    assert yt1 == pytest.approx(yt2)

    a1, b1, isOK1 = conv._tangentToPositioner(
        xt1, yt1, xBetas, yBetas, la, alphaOff, betaOff
    )

    a2, b2, isOK2 = conv.tangentToPositioner(
        xt1, yt1, xBetas, yBetas, la, alphaOff, betaOff
    )

    assert a1 == pytest.approx(a2)
    assert b1 == pytest.approx(b2)
    assert isOK1 == isOK2
    assert isOK1

    assert a1 == pytest.approx(alphas)
    assert b1 == pytest.approx(betas)


def test_tangentAndPositionerSafeArr():
    ## if you're not near the edges of travel,
    # things shouldn't be weird?
    nPts = 5
    alphas = numpy.random.uniform(10,350, nPts)
    betas = numpy.random.uniform(10, 170, nPts)
    yBetas = numpy.random.uniform(-0.5,0.5, nPts)
    xBetas = [15]*nPts
    la = 7.4
    alphaOff = numpy.random.uniform(-1,1)
    betaOff = numpy.random.uniform(-1,1)

    for ii in [1,2]:
        # non-array
        xt1, yt1 = conv._positionerToTangent(
            alphas[ii], betas[ii], xBetas[ii], yBetas[ii], la, alphaOff, betaOff
        )

        xt2, yt2 = conv.positionerToTangent(
            alphas[ii], betas[ii], xBetas[ii], yBetas[ii], la, alphaOff, betaOff
        )


        assert xt1 == pytest.approx(xt2)
        assert yt1 == pytest.approx(yt2)


        a1, b1, isOK1 = conv._tangentToPositioner(
            xt1, yt1, xBetas[ii], yBetas[ii], la, alphaOff, betaOff
        )

        a2, b2, isOK2 = conv.tangentToPositioner(
            xt1, yt1, xBetas[ii], yBetas[ii], la, alphaOff, betaOff
        )


        assert a1 == pytest.approx(a2)
        assert b1 == pytest.approx(b2)
        assert isOK1 == isOK2
        assert isOK1

        assert a1 == pytest.approx(alphas[ii])
        assert b1 == pytest.approx(betas[ii])

    # array
    xt1, yt1 = conv._positionerToTangent(
        alphas, betas, xBetas, yBetas, la, alphaOff, betaOff
    )

    xt2, yt2 = conv.positionerToTangent(
        alphas, betas, xBetas, yBetas, la, alphaOff, betaOff
    )


    assert xt1 == pytest.approx(xt2)
    assert yt1 == pytest.approx(yt2)


    a1, b1, isOK1 = conv._tangentToPositioner(
        xt1, yt1, xBetas, yBetas, la, alphaOff, betaOff
    )

    a2, b2, isOK2 = conv.tangentToPositioner(
        xt1, yt1, xBetas, yBetas, la, alphaOff, betaOff
    )


    # import pdb; pdb.set_trace()
    assert a1 == pytest.approx(a2)
    assert b1 == pytest.approx(b2)
    assert not False in isOK1
    assert not False in isOK2

    assert a1 == pytest.approx(alphas)
    assert b1 == pytest.approx(betas)


def test_unreachableTangent():


    r = 24
    thetas = numpy.random.uniform(0, 2*numpy.pi,20)
    xts = r*numpy.cos(thetas)
    yts = r*numpy.sin(thetas)

    for fx,fy in [modelMetXY, modelBossXY, modelApXY]:
        a1,b1,isOK1 = conv._tangentToPositioner(
            xts,yts,fx,fy,la
        )

        a2,b2,isOK2 = conv.tangentToPositioner(
            xts,yts,fx,fy,la
        )
        assert not True in isOK1
        assert not True in isOK2


def plot_wholeRange():
    # not really a test
    nPts = 100000
    rs = numpy.random.uniform(0, (la+17)**2, nPts)
    thetas = numpy.random.uniform(0, 2*numpy.pi, nPts)
    xts = numpy.sqrt(rs)*numpy.cos(thetas)
    yts = numpy.sqrt(rs)*numpy.sin(thetas)

    ii = 0
    for fx, fy in [modelMetXY, modelBossXY, modelApXY]:

        a1,b1,isOK1 = conv._tangentToPositioner(
            xts,yts,fx,fy,la
        )

        a2,b2,isOK2 = conv.tangentToPositioner(
            xts,yts,fx,fy,la
        )

        fig, axs = plt.subplots(1,2, figsize=(13,8))

        sns.scatterplot(x=xts, y=yts, hue=isOK1, alpha=0.2, ax=axs[0])
        sns.scatterplot(x=xts, y=yts, hue=isOK2, alpha=0.2, ax=axs[1])
        axs[0].set_aspect("equal")
        axs[1].set_aspect("equal")
        fig.suptitle("iter %i"%ii)


        _a1 = a1[isOK1]
        _b1 = b1[isOK1]
        xts1 = xts[isOK1]
        yts1 = yts[isOK1]

        _a2 = a2[isOK2]
        _b2 = b2[isOK2]
        xts2 = xts[isOK2]
        yts2 = yts[isOK2]

        fig, axs = plt.subplots(2,2, figsize=(10,10))
        axs = axs.flatten()

        sns.scatterplot(x=xts1, y=yts1, hue=_a1, alpha=0.2, ax=axs[0])
        sns.scatterplot(x=xts1, y=yts1, hue=_b1, alpha=0.2, ax=axs[1])
        sns.scatterplot(x=xts2, y=yts2, hue=_a2, alpha=0.2, ax=axs[2])
        sns.scatterplot(x=xts2, y=yts2, hue=_b2, alpha=0.2, ax=axs[3])
        axs[0].set_aspect("equal")
        axs[1].set_aspect("equal")
        axs[2].set_aspect("equal")
        axs[3].set_aspect("equal")
        axs[0].set_title("alpha")
        axs[1].set_title("beta")

        fig.suptitle("iter %i"%ii)
        ii += 1
    plt.show()


def plot_degenerateSolns():

    npts = 100000
    alphas = numpy.random.uniform(0, 359.9999, npts)
    betas = numpy.random.uniform(0, 180, npts)


    for fx, fy in [modelMetXY, modelBossXY, modelApXY]:
        tx, ty = conv.positionerToTangent(
            alphas, betas, fx, fy, la
        )

        _alphas, _betas, isOK = conv.tangentToPositioner(
            tx, ty, fx, fy, la
        )
        da = alphas - _alphas
        ida = numpy.abs(da) > 1e-5
        db = betas - _betas
        idb = numpy.abs(db) > 1e-5
        fig = plt.figure()
        sns.scatterplot(x=tx[ida], y=ty[ida], hue=da[ida], s=3, alpha=1)
        plt.axis("equal")
        plt.title("d alpha")
        plt.show()

        fig = plt.figure()
        sns.scatterplot(x=tx[idb], y=ty[idb], hue=db[idb], s=3, alpha=1)
        plt.axis("equal")
        plt.title("d beta")
        plt.show()


def test_wokAndTangent():

    for holeID in holeIDs:
        row = calibration.wokCoords.loc[('APO', holeID)]
        b = [round(row.xWok, 5),
             round(row.yWok, 5),
             round(row.zWok, 5)]
        iHat = [float(row.ix), float(row.iy), float(row.iz)]
        jHat = [float(row.jx), float(row.jy), float(row.jz)]
        kHat = [float(row.kx), float(row.ky), float(row.kz)]

        for tx,ty,tz in tanCoordList:
            scaleFac = numpy.random.uniform(0.9,1.1)
            dx = numpy.random.uniform(-0.01,0.01)
            dy = numpy.random.uniform(-0.01,0.01)
            dz = numpy.random.uniform(-0.01,0.01)

            for addMe in [0, numpy.zeros(4)]:
                _tx = tx + addMe
                _ty = ty + addMe
                _tz = tz + addMe

                wx1,wy1,wz1 = conv.tangentToWok(_tx, _ty, _tz, b, iHat, jHat, kHat,
                    scaleFac=scaleFac, dx=dx, dy=dy, dz=dz)
                wx2,wy2,wz2 = conv._tangentToWok(_tx, _ty, _tz, b, iHat, jHat, kHat,
                    scaleFac=scaleFac, dx=dx, dy=dy, dz=dz)

                assert wx1 == pytest.approx(wx2)
                assert wy1 == pytest.approx(wy2)
                assert wz1 == pytest.approx(wz2)

                tx1,ty1,tz1 = conv.wokToTangent(
                    wx1, wy1, wz1, b, iHat, jHat, kHat,
                    scaleFac=scaleFac, dx=dx, dy=dy, dz=dz)
                tx2,ty2,tz2 = conv._wokToTangent(wx1, wy1, wz1, b, iHat, jHat, kHat,
                    scaleFac=scaleFac, dx=dx, dy=dy, dz=dz)

                assert tx1 == pytest.approx(tx2)
                assert ty1 == pytest.approx(ty2)
                assert tz1 == pytest.approx(tz2)

                assert _tx == pytest.approx(tx2)
                assert _ty == pytest.approx(ty2)
                assert _tz == pytest.approx(tz2)

        # break

def test_lefthand():
    npts = 5
    alphaOff = 0
    betaOff = 0
    la = 7.4
    xBeta = 15
    yBeta = 0

    alphas = numpy.random.uniform(0, 360, npts)
    betas = numpy.random.uniform(0, 180, npts)


    xt, yt = conv.positionerToTangent(
        alphas, betas, xBeta, yBeta, la, alphaOff, betaOff
    )

    # lefthand version
    _alphaLeft, _betaLeft, isOK = conv.tangentToPositioner(
        xt, yt, xBeta, yBeta, la, lefthand=True
    )

    # righthand version
    # lefthand version
    _alphaRight, _betaRight, isOK = conv.tangentToPositioner(
        xt, yt, xBeta, yBeta, la, lefthand=False
    )

    assert numpy.all(_betaLeft > 180)
    assert numpy.all(_betaRight < 180)
    # print(xt, yt)

if __name__ == "__main__":
    # plot_degenerateSolns()
    test_lefthand()

#     tx,ty,tz = numpy.zeros(1000)+22, numpy.zeros(1000)-4, numpy.zeros(1000)



#     import pdb; pdb.set_trace()

#     tstart = time.time()
#     _tx1, _ty1, _tz1 = conv.wokToTangent(wx,wy,wz,b,iHat,jHat,kHat)
#     print("carr", time.time()-tstart)
#     tstart = time.time()
#     _tx2, _ty2, _tz2 = conv._wokToTangent(wx,wy,wz,b,iHat,jHat,kHat)
#     print("narr", time.time()-tstart)


#     import pdb; pdb.set_trace()
#     break


# import pdb; pdb.set_trace()



import numpy

from coordio.conv import (fieldToFocal, focalToWok, proj2XYplane,
                          sph2Cart, cart2Sph, fieldAngle2Cart, cart2FieldAngle)

SMALL_NUM = 2e-9
nPts = 100000
thetas = numpy.random.uniform(size=nPts) * 2 * numpy.pi
# maximum of 4 degrees off axis
phis = numpy.radians(numpy.random.uniform(size=nPts) * 4)
r = 1
xs = r * numpy.cos(thetas) * numpy.sin(phis)
ys = r * numpy.sin(thetas) * numpy.sin(phis)
zs = r * numpy.cos(phis)


def test_phiConversion(verbose=False):
    for fieldAngle in numpy.linspace(-10,10,100):
        x,y,z = fieldAngle2Cart(fieldAngle, 0)
        theta, phi = cart2Sph(x,y,z)
        if verbose:
            print(numpy.abs(phi-fieldAngle))
        else:
            assert numpy.abs(phi-numpy.abs(fieldAngle)) < SMALL_NUM

        x,y,z = fieldAngle2Cart(0, fieldAngle)
        theta, phi = cart2Sph(x,y,z)
        if verbose:
            print(numpy.abs(phi-fieldAngle))
        else:
            assert numpy.abs(phi-numpy.abs(fieldAngle)) < SMALL_NUM


def test_cartField():
    # pick some random points on the unit sphere near the +Z cap
    for theta, phi, x, y, z in zip(thetas, phis, xs, ys, zs):
        xField, yField = cart2FieldAngle(x, y, z)
        if theta < numpy.pi / 2 or theta > 3 * numpy.pi / 2:
            assert xField < 0
        else:
            assert xField > 0

        if theta < numpy.pi:
            assert yField < 0
        else:
            assert yField > 0

        xSolve, ySolve, zSolve = fieldAngle2Cart(xField, yField)
        # print(xSolve - x, ySolve - y, zSolve - z)

        assert numpy.abs(xSolve-x) < SMALL_NUM
        assert numpy.abs(ySolve-y) < SMALL_NUM
        assert numpy.abs(zSolve-z) < SMALL_NUM


def test_cartFieldCycle():
    # check round trippage
    inds = numpy.random.choice(range(nPts), size=100)
    for ind in inds:
        x = xs[ind]
        y = ys[ind]
        z = zs[ind]
        _x = xs[ind]
        _y = ys[ind]
        _z = zs[ind]
        for ii in range(100):
            # print("ii", ii)
            xField, yField = cart2FieldAngle(_x,_y,_z)
            _x, _y, _z = fieldAngle2Cart(xField, yField)
            # repeated round trips require extra numerical buffer
            assert numpy.abs(_x-x) < SMALL_NUM
            assert numpy.abs(_y-y) < SMALL_NUM
            assert numpy.abs(_z-z) < SMALL_NUM


def test_sphCartCycle():
    # check round trippage
    inds = numpy.random.choice(range(nPts), size=100)
    for ind in inds:
        x = xs[ind]
        y = ys[ind]
        z = zs[ind]
        theta = numpy.degrees(thetas[ind])
        phi = numpy.degrees(phis[ind])
        _x = xs[ind]
        _y = ys[ind]
        _z = zs[ind]
        for ii in range(100):
            _theta, _phi = cart2Sph(_x, _y, _z)
            # print("[%.4f, %.4f] == %.5e, %.5e"%(theta, phi, phiTheta[0]-theta, phiTheta[1]-phi))
            assert numpy.abs(_theta - theta) < SMALL_NUM
            assert numpy.abs(_phi - phi) < SMALL_NUM
            _x, _y, _z = sph2Cart(_theta, _phi)
            assert numpy.abs(_x-x) < SMALL_NUM
            assert numpy.abs(_y-y) < SMALL_NUM
            assert numpy.abs(_z-z) < SMALL_NUM

def test_fieldToFocal():
    # test zero field
    for obs in ["APO", "LCO"]:
        for waveCat in ["Apogee", "Boss", "GFA"]:
            x,y,z = 0,0,1
            theta, phi = cart2Sph(x,y,z)
            # check on axis
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert numpy.abs(_x) < SMALL_NUM
            assert numpy.abs(_y) < SMALL_NUM
            # check array on axis
            x = [0,0,0]
            y = [0,0,0]
            z = [1,1,1]
            theta, phi = cart2Sph(x,y,z)
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert numpy.max(numpy.abs(_x)) < SMALL_NUM
            assert numpy.max(numpy.abs(_y)) < SMALL_NUM

            theta = 0
            phi = 0.5 # 0.5 degrees off axis
            x,y,z = sph2Cart(theta, phi)
            assert x > 0
            assert numpy.abs(y) < SMALL_NUM
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert _x > 0
            assert numpy.abs(_y) < SMALL_NUM
            assert _z < 0

            theta = 90
            phi = 0.5 # 0.5 degrees off axis
            x,y,z = sph2Cart(theta, phi)
            assert numpy.abs(x) < SMALL_NUM
            assert y > 0
            assert z > 0
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert _y > 0
            assert numpy.abs(_x) < SMALL_NUM
            assert _z < 0

            theta = 180
            phi = 0.5 # 0.5 degrees off axis
            x,y,z = sph2Cart(theta, phi)
            assert numpy.abs(y) < SMALL_NUM
            assert x < 0
            assert z > 0
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert _x < 0
            assert numpy.abs(_y) < SMALL_NUM
            assert _z < 0

            theta = 270
            phi = 0.5 # 0.5 degrees off axis
            x,y,z = sph2Cart(theta, phi)
            assert numpy.abs(x) < SMALL_NUM
            assert y < 0
            assert z > 0
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert _y < 0
            assert numpy.abs(_x) < SMALL_NUM
            assert _z < 0

            theta = 300
            phi = 0.5 # 0.5 degrees off axis
            x,y,z = sph2Cart(theta, phi)
            assert x > 0
            assert y < 0
            assert z > 0
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert _y < 0
            assert _x > 0
            assert _z < 0

            theta = 380
            phi = 0.5 # 0.5 degrees off axis
            x,y,z = sph2Cart(theta, phi)
            assert x > 0
            assert y > 0
            assert z > 0
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert _y > 0
            assert _x > 0
            assert _z < 0

            theta = -20
            phi = 0.5 # 0.5 degrees off axis
            x,y,z = sph2Cart(theta, phi)
            assert x > 0
            assert y < 0
            assert z > 0
            _x,_y,_z,R,b,w = fieldToFocal(theta, phi, obs, waveCat)
            assert _y < 0
            assert _x > 0
            assert _z < 0


def test_focalToWok():
    zOff = -100
    xOff = 0
    yOff = 0
    xTilt = 0
    yTilt = 0
    positionAngle = 0

    xFocal, yFocal, zFocal = 0, 0, 0
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )

    assert xWok == 0
    assert yWok == 0
    assert zWok == -1 * zOff

    positionAngle = 90  # +y wok aligned with +x FP

    xFocal, yFocal, zFocal = 10, 0, 0

    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )
    assert yWok == xFocal
    assert numpy.abs(xWok) < SMALL_NUM

    positionAngle = -90 # +y wok aligned with -x FP

    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )

    assert yWok == -1*xFocal
    assert numpy.abs(xWok) < SMALL_NUM

    # test translation
    xFocal, yFocal, zFocal = 0, 0, 0
    xOff = 10
    positionAngle = 0
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )

    assert xWok == -1*xOff
    assert numpy.abs(yWok) < SMALL_NUM

    yOff = 10
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )

    assert xWok == -1*xOff
    assert yWok == -1*yOff

    positionAngle = 45
    xOff, yOff = 10, 10
    xFocal, yFocal, zFocal = 0,0,0
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )
    assert xWok == yWok
    assert xWok < 0

    positionAngle = 45
    xOff, yOff = -10, 10
    xFocal, yFocal, zFocal = 0,0,0
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )
    assert xWok == -1*yWok
    assert xWok > 0

    b = 0.5*numpy.sqrt(2*10**2)
    a = numpy.sqrt(2*b**2)
    positionAngle = 45
    xOff, yOff, zOff = 10, 10, 0
    xTilt, yTilt = 0, 0
    xFocal, yFocal, zFocal = b, b, 0

    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )
    assert numpy.abs(xWok + a) < SMALL_NUM
    assert numpy.abs(yWok) < SMALL_NUM

    # test tilts
    positionAngle =0
    xOff, yOff, zOff = 0, 0, 0
    xFocal, yFocal, zFocal = 1, 0, 0
    xTilt = 1
    yTilt = 0
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )
    assert xWok == xFocal
    assert yWok == yFocal
    assert zWok == zFocal

    positionAngle =0
    xOff, yOff, zOff = 0, 0, 0
    xFocal, yFocal, zFocal = 0, 1, 0
    xTilt = 1
    yTilt = 0
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )

    assert xWok == 0
    assert yWok < 1
    assert yWok > 0
    assert zWok < 0

    positionAngle =0
    xOff, yOff, zOff = 0, 0, 0
    xFocal, yFocal, zFocal = 1, 1, 0
    xTilt = 1
    yTilt = 4
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )

    assert xWok > 0
    assert xWok < 1
    assert yWok < 1
    assert yWok > 0
    assert yWok < 1
    assert zWok > 0

    positionAngle =0
    xOff, yOff, zOff = 0, 0, 0
    xFocal, yFocal, zFocal = 1, 1, 0
    xTilt = 1
    yTilt = -1
    xWok, yWok, zWok = focalToWok(
        xFocal, yFocal, zFocal, positionAngle,
        xOff, yOff, zOff, xTilt, yTilt,
        fpScale=1, projectFlat=False
    )

    assert xWok > 0
    assert xWok < 1
    assert yWok < 1
    assert yWok > 0
    assert yWok < 1
    assert zWok < 0


def test_xyProj():
    rayOrigin = [0,0,1]
    r = 0.5
    thetas = numpy.linspace(-numpy.pi, numpy.pi) # put in arctan2 domain
    x = r*numpy.cos(thetas)
    y = r*numpy.sin(thetas)
    z = [0.5]*len(x)
    px, py, pz, ps = proj2XYplane(x,y,z, rayOrigin)
    # print(ps)
    mags = numpy.sqrt(px**2+py**2)
    assert numpy.max(numpy.abs(mags-1)) < SMALL_NUM
    assert numpy.max(numpy.abs(pz)) < SMALL_NUM

    _thetas = numpy.arctan2(py, px)
    assert numpy.max(numpy.abs(_thetas-thetas)) < SMALL_NUM

    r = 1.5
    x = r*numpy.cos(thetas)
    y = r*numpy.sin(thetas)
    z = [-0.5]*len(x)
    _px, _py, _pz, _ps = proj2XYplane(x,y,z, rayOrigin)

    assert numpy.max(numpy.abs(px-_px)) < SMALL_NUM
    assert numpy.max(numpy.abs(py-_py)) < SMALL_NUM
    assert numpy.max(numpy.abs(_pz)) < SMALL_NUM

    _thetas = numpy.arctan2(_py, _px)
    assert numpy.max(numpy.abs(_thetas-thetas)) < SMALL_NUM

    rayOrigin = [0.5, 0.5, 100]
    x = 0.5
    y = 0.5
    z = 1
    px, py, pz, ps = proj2XYplane(x,y,z, rayOrigin)
    assert numpy.abs(px-x) < SMALL_NUM
    assert numpy.abs(py-y) < SMALL_NUM
    assert numpy.abs(pz) < SMALL_NUM

    rayOrigin = [-1, 0, 2]
    x = 0
    y = 0
    z = 1

    px, py, pz, ps = proj2XYplane(x,y,z, rayOrigin)
    assert numpy.abs(px-1) < SMALL_NUM
    assert numpy.abs(py) < SMALL_NUM
    assert numpy.abs(pz) < SMALL_NUM


def test_xyProj2():
    xExpect = 1
    yExpect = 0
    zExpect = 0
    projExpect = numpy.sqrt(2*0.5**2)
    rayOrigin = numpy.array([0,0,1])
    x, y, z = 0.5, 0, 0.5
    px, py, pz, pd = proj2XYplane(x,y,z, rayOrigin)

    assert numpy.abs(xExpect-px) < SMALL_NUM
    assert numpy.abs(yExpect-py) < SMALL_NUM
    assert numpy.abs(zExpect-pz) < SMALL_NUM
    assert numpy.abs(projExpect-pd) < SMALL_NUM


    x, y, z = 1.5, 0, -0.5
    px, py, pz, pd = proj2XYplane(x,y,z, rayOrigin)

    assert numpy.abs(xExpect-px) < SMALL_NUM
    assert numpy.abs(yExpect-py) < SMALL_NUM
    assert numpy.abs(zExpect-pz) < SMALL_NUM
    assert numpy.abs(-1*projExpect-pd) < SMALL_NUM

    x = [0.5, 1.5]
    y = [0, 0]
    z = [0.5, -0.5]

    px, py, pz, pd = proj2XYplane(x,y,z, rayOrigin)

    assert numpy.abs(xExpect - px[0]) < SMALL_NUM
    assert numpy.abs(yExpect - py[0]) < SMALL_NUM
    assert numpy.abs(zExpect - pz[0]) < SMALL_NUM
    assert numpy.abs(projExpect - pd[0]) < SMALL_NUM

    assert numpy.abs(xExpect - px[1]) < SMALL_NUM
    assert numpy.abs(yExpect - py[1]) < SMALL_NUM
    assert numpy.abs(zExpect - pz[1]) < SMALL_NUM
    assert numpy.abs(-1 * projExpect - pd[1]) < SMALL_NUM
